(*  Title:       AbstractCalculi.thy
    Author:      Stephan van Staden, ETH Zurich, Switzerland
    Description: Abstract calculi for deductive and operational reasoning.
    Copyright    2012 ETH Zurich
*)

theory AbstractCalculi
  imports Languages
begin

declare [[ smt_solver = remote_z3]]
declare [[ smt_timeout = 60 ]]
declare [[ z3_options = "-memory:500" ]]

(* Hoare logic *)

definition  hoare :: "'a language \<Rightarrow> 'a language \<Rightarrow> 'a language \<Rightarrow> bool" ("_\<lbrace>_\<rbrace>_" [54,54,54] 53)
  where "P\<lbrace>Q\<rbrace>R \<equiv> P;Q \<subseteq> R"

theorem Hskip: "P\<lbrace>skip\<rbrace>P"
  by (unfold hoare_def, metis order_refl lang_concat_unit2)
theorem Hseq: "P\<lbrace>Q\<rbrace>R \<Longrightarrow> R\<lbrace>Q'\<rbrace>S \<Longrightarrow> P\<lbrace>Q;Q'\<rbrace>S"
  by (unfold hoare_def, metis lang_concat_associative lang_concat_monotone1 order_trans)
theorem Hchoice: "(\<forall>Q\<in>X.  P\<lbrace>Q\<rbrace>R) \<Longrightarrow> P\<lbrace>\<Union>X\<rbrace>R"
proof -
  assume "\<forall>Q\<in>X.  P\<lbrace>Q\<rbrace>R"
  then have "\<Union>{P'. \<exists>Q \<in> X. P' = P;Q} \<subseteq> R" by (unfold hoare_def, auto)
  then have "P;\<Union>{P'. \<exists>Q \<in> X. P' = Q} \<subseteq> R" by (simp add: lang_concat_distr_Union2)
  then have "P;\<Union>X \<subseteq> R" by auto
  then show "P\<lbrace>\<Union>X\<rbrace>R" by (simp add: hoare_def)
qed
theorem Hiter: "P\<lbrace>Q\<rbrace>P \<Longrightarrow> P\<lbrace>Q\<^sup>\<star>\<rbrace>P"
  by (unfold hoare_def, metis sup_commute kleene4 le_iff_sup order_refl)
theorem Hcons: "P' \<subseteq> P \<Longrightarrow> P\<lbrace>Q\<rbrace>R \<Longrightarrow> R \<subseteq> R' \<Longrightarrow> P'\<lbrace>Q\<rbrace>R'"
  by (unfold hoare_def, metis order_trans lang_concat_monotone1)
theorem Hdisj: "(\<forall>P\<in>X.  P\<lbrace>Q\<rbrace>R) \<Longrightarrow> \<Union>X\<lbrace>Q\<rbrace>R"
proof -
  assume "\<forall>P\<in>X.  P\<lbrace>Q\<rbrace>R"
  then have "\<Union>{P'. \<exists>P \<in> X. P' = P;Q} \<subseteq> R" by (unfold hoare_def, auto)
  then have "\<Union>{P'. \<exists>P \<in> X. P' = P};Q \<subseteq> R" by (simp add: lang_concat_distr_Union1)
  then have "\<Union>X;Q \<subseteq> R" by auto
  then show "\<Union>X\<lbrace>Q\<rbrace>R" by (simp add: hoare_def)
qed
theorem Hconj: "P\<lbrace>Q\<rbrace>R \<Longrightarrow> P'\<lbrace>Q'\<rbrace>R' \<Longrightarrow> P\<inter>P'\<lbrace>Q\<inter>Q'\<rbrace>R\<inter>R'"
proof -
  assume "P\<lbrace>Q\<rbrace>R"
  then have "(P\<inter>P');Q \<subseteq> R" by (metis hoare_def inf_le1 lang_concat_monotone1 subset_trans)
  then have a: "(P\<inter>P');(Q\<inter>Q') \<subseteq> R" by (metis inf_le1 lang_concat_monotone2 order_trans)
  assume "P'\<lbrace>Q'\<rbrace>R'"
  then have "(P\<inter>P');Q' \<subseteq> R'" by (metis hoare_def inf_le2 lang_concat_monotone1 subset_trans)
  then have b: "(P\<inter>P');(Q\<inter>Q') \<subseteq> R'" by (metis inf_le2 lang_concat_monotone2 order_trans)
  from a b show "P\<inter>P'\<lbrace>Q\<inter>Q'\<rbrace>R\<inter>R'" by (metis hoare_def inf_greatest)
qed
theorem Hframe: "P\<lbrace>Q\<rbrace>R \<Longrightarrow> F\<parallel>P\<lbrace>Q\<rbrace>F\<parallel>R"
  by (unfold hoare_def, metis lang_interleave_monotone2 order_trans small_exchange2)
theorem Hconc: "P\<lbrace>Q\<rbrace>R \<Longrightarrow> P'\<lbrace>Q'\<rbrace>R' \<Longrightarrow> P\<parallel>P'\<lbrace>Q\<parallel>Q'\<rbrace>R\<parallel>R'"
  by (unfold hoare_def, metis lang_interleave_monotone2 lang_interleave_monotone1 order_trans exchange)

(* Small-step operational calculi. *)

locale abstract_small_step_calculi =
  fixes Actions :: "'a language set"
  assumes skip_in_Actions: "skip \<in> Actions"
begin
definition a_plotkin :: "'a language \<Rightarrow> 'a language \<Rightarrow> 'a language \<Rightarrow> 'a language \<Rightarrow> bool" ("<_,_>\<rightarrow>\<^isub>a<_,_>" [54,54,54,54] 53) where
  "<P,s>\<rightarrow>\<^isub>a<P',s'> \<equiv> \<exists>Q. Q \<in> Actions \<and> P \<supseteq> Q;P' \<and> s;Q \<supseteq> s'"
lemma Paction: "P \<in> Actions \<Longrightarrow> s' \<subseteq> s;P \<Longrightarrow> <P,s>\<rightarrow>\<^isub>a<skip,s'>"
  by (metis a_plotkin_def lang_concat_unit2 subset_refl)
lemma Pseq1: "<skip;P,s>\<rightarrow>\<^isub>a<P,s>"
  by (metis a_plotkin_def equalityD1 lang_concat_unit2 skip_in_Actions)
lemma Pseq2: "<P,s>\<rightarrow>\<^isub>a<R,s'> \<Longrightarrow> <P;P',s>\<rightarrow>\<^isub>a<R;P',s'>"
  by (metis a_plotkin_def lang_concat_associative lang_concat_monotone1)
lemma Pchoice: "P \<in> X \<Longrightarrow> <\<Union>X,s>\<rightarrow>\<^isub>a<P,s>"
  by (metis Collect_def Pow_def a_plotkin_def in_mono lang_concat_unit1 lang_concat_unit2 mem_def skip_in_Actions subset_Pow_Union subset_refl)
lemma Piter1: "<P\<^sup>\<star>,s>\<rightarrow>\<^isub>a<skip,s>"
  by (metis a_plotkin_def lang_concat_unit2 skip_in_Actions skip_included_in_kleene_star subset_refl)
lemma Piter2: "<P\<^sup>\<star>,s>\<rightarrow>\<^isub>a<P;P\<^sup>\<star>,s>"
  by (metis a_plotkin_def kleene1part lang_concat_unit1 lang_concat_unit2 skip_in_Actions subset_refl)
lemma Pconc1: "<skip\<parallel>P,s>\<rightarrow>\<^isub>a<P,s>"
  by (metis Pseq1 lang_concat_unit1 lang_interleave_unit1)
lemma Pconc2: "<P\<parallel>skip,s>\<rightarrow>\<^isub>a<P,s>"
  by (metis Pconc1 lang_interleave_commutative)
lemma Pconc3: "<P,s>\<rightarrow>\<^isub>a<R,s'> \<Longrightarrow> <P\<parallel>P',s>\<rightarrow>\<^isub>a<R\<parallel>P',s'>"
proof -
  assume "<P,s>\<rightarrow>\<^isub>a<R,s'>"
  then have "\<exists>Q. Q \<in> Actions \<and> P \<supseteq> Q;R \<and> s;Q \<supseteq> s'" by (unfold a_plotkin_def)
  then obtain Q where a: "Q \<in> Actions \<and> P \<supseteq> Q;R \<and> s;Q \<supseteq> s'" by blast
  from a have "Q \<in> Actions \<and> P\<parallel>P' \<supseteq> Q;(R\<parallel>P') \<and> s;Q \<supseteq> s'" by (metis lang_interleave_commutative lang_interleave_monotone2 small_exchange1 xt1(6))
  then show "<P\<parallel>P',s>\<rightarrow>\<^isub>a<R\<parallel>P',s'>" by (metis a_plotkin_def)
qed
lemma Pconc4: "<P,s>\<rightarrow>\<^isub>a<R,s'> \<Longrightarrow> <P'\<parallel>P,s>\<rightarrow>\<^isub>a<P'\<parallel>R,s'>"
  by (metis Pconc3 lang_interleave_commutative)
definition  a_milner :: "'a language \<Rightarrow> 'a language \<Rightarrow> 'a language \<Rightarrow> bool" ("_ -_\<rightarrow>\<^isub>a _" [54,54,54] 53) where
  "P -Q\<rightarrow>\<^isub>a R \<equiv> Q \<in> Actions \<and> P \<supseteq> Q;R"
lemma Maction: "P \<in> Actions \<Longrightarrow> P -P\<rightarrow>\<^isub>a skip"
  by (metis a_milner_def lang_concat_unit2 order_refl)
lemma Mseq1: "skip;P -skip\<rightarrow>\<^isub>a P"
  by (metis a_milner_def equalityD1 skip_in_Actions)
lemma Mseq2: "P -Q\<rightarrow>\<^isub>a R \<Longrightarrow> P;P' -Q\<rightarrow>\<^isub>a R;P'"
  by (metis a_milner_def lang_concat_associative lang_concat_monotone1)
lemma Mchoice: "P \<in> X \<Longrightarrow> \<Union>X -skip\<rightarrow>\<^isub>a P"
  by (metis Sup_upper a_milner_def lang_concat_unit1 skip_in_Actions)
lemma Miter1: "P\<^sup>\<star> -skip\<rightarrow>\<^isub>a skip"
  by (metis a_milner_def lang_concat_unit2 skip_in_Actions skip_included_in_kleene_star)
lemma Miter2: "P\<^sup>\<star> -skip\<rightarrow>\<^isub>a P;P\<^sup>\<star>"
  by (metis a_milner_def kleene1part lang_concat_unit1 skip_in_Actions)
lemma Mconc1: "skip\<parallel>P -skip\<rightarrow>\<^isub>a P"
  by (metis Mseq1 lang_concat_unit1 lang_interleave_unit1)
lemma Mconc2: "P\<parallel>skip -skip\<rightarrow>\<^isub>a P"
  by (metis Mconc1 lang_interleave_commutative)
lemma Mconc3: "P -Q\<rightarrow>\<^isub>a R \<Longrightarrow> P\<parallel>P' -Q\<rightarrow>\<^isub>a R\<parallel>P'"
  by (metis (no_types) a_milner_def lang_interleave_monotone1 small_exchange1 subset_trans)
lemma Mconc4: "P -Q\<rightarrow>\<^isub>a R \<Longrightarrow> P'\<parallel>P -Q\<rightarrow>\<^isub>a P'\<parallel>R"
  by (metis Mconc3 lang_interleave_commutative)
end

(* Kahn calculus *)

definition  a_kahn :: "'a language \<Rightarrow> 'a language \<Rightarrow> 'a language \<Rightarrow> bool" ("<_,_>\<rightarrow>\<^isub>a_" [54,54,54] 53)
  where "<P,s>\<rightarrow>\<^isub>as' \<equiv> s' \<subseteq> s;P"

lemma  Kimprovement: "P \<subseteq> P' \<Longrightarrow> <P,s>\<rightarrow>\<^isub>as' \<Longrightarrow> <P',s>\<rightarrow>\<^isub>as'"
  by (unfold a_kahn_def, metis order_trans lang_concat_monotone2)

theorem Kskip: "<skip,s>\<rightarrow>\<^isub>as"
  by (unfold a_kahn_def, metis order_refl lang_concat_unit2)
theorem Kseq: "<P,s>\<rightarrow>\<^isub>as' \<Longrightarrow> <P',s'>\<rightarrow>\<^isub>as'' \<Longrightarrow> <P;P',s>\<rightarrow>\<^isub>as''"
  by (unfold a_kahn_def, metis order_trans lang_concat_associative lang_concat_monotone1)
theorem Kchoice: "P \<in> X \<Longrightarrow> <P,s>\<rightarrow>\<^isub>as' \<Longrightarrow> <\<Union>X,s>\<rightarrow>\<^isub>as'"
  by (metis Kimprovement Union_upper)
theorem Kiter1: "<P\<^sup>\<star>,s>\<rightarrow>\<^isub>as"
  by (metis Kimprovement Kskip skip_included_in_kleene_star)
theorem Kiter2: "<P,s>\<rightarrow>\<^isub>as' \<Longrightarrow> <P\<^sup>\<star>,s'>\<rightarrow>\<^isub>as'' \<Longrightarrow> <P\<^sup>\<star>,s>\<rightarrow>\<^isub>as''"
proof -
  assume a: "<P,s>\<rightarrow>\<^isub>as'"
  assume b: "<P\<^sup>\<star>,s'>\<rightarrow>\<^isub>as''"
  from a b have c: "<P;P\<^sup>\<star>,s>\<rightarrow>\<^isub>as''" by (rule Kseq)
  from kleene1part c show "<P\<^sup>\<star>,s>\<rightarrow>\<^isub>as''" by (rule Kimprovement)
qed
theorem Kconc1: "<P,s>\<rightarrow>\<^isub>as' \<Longrightarrow> <P',s'>\<rightarrow>\<^isub>as'' \<Longrightarrow> <P\<parallel>P',s>\<rightarrow>\<^isub>as''"
  by (metis Kimprovement Kseq interleave_contains_concat)
theorem Kconc2: "<P',s>\<rightarrow>\<^isub>as' \<Longrightarrow> <P,s'>\<rightarrow>\<^isub>as'' \<Longrightarrow> <P\<parallel>P',s>\<rightarrow>\<^isub>as''"
  by (metis Kconc1 lang_interleave_commutative)

end

