(*  Title:       DeductiveCalculi.thy
    Author:      Stephan van Staden, ETH Zurich, Switzerland
    Description: The deductive calculi for descriptions and commands that use views.
    Copyright    2012 ETH Zurich
*)

theory DeductiveCalculi
  imports AbstractCalculi DescriptionsAndCommands
begin

declare [[ smt_solver = remote_z3]]
declare [[ smt_timeout = 60 ]]
declare [[ z3_options = "-memory:500" ]]

locale views_symbols =
  fixes entailment :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<Turnstile>" 50)
  and Join :: "'a set \<Rightarrow> 'a" ("\<Squnion>_" [900] 900)
  and Meet :: "'a set \<Rightarrow> 'a" ("\<Sqinter>_" [900] 900)
  and star :: "'a \<Rightarrow> 'a \<Rightarrow> 'a" (infixl "\<star>" 80)
  and unitview :: "'a"
  and angelic_preorder :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<prec>" 50)
  and erase :: "'a \<Rightarrow> 'b set" ("\<lfloor>_\<rfloor>" [70] 900)
  and Axioms :: "('a \<times> 'b atom \<times> 'a) set"

locale views = views_symbols +
  views_lattice: complete_lattice Meet Join "\<lambda>v v'. \<Sqinter>{v,v'}" "op \<Turnstile>" "\<lambda>v v'. v \<Turnstile> v' \<and> \<not> v' \<Turnstile> v" "\<lambda>v v'. \<Squnion>{v,v'}" "\<Squnion>{}" "\<Sqinter>{}" +
  views_monoid: comm_monoid "op \<star>" unitview +
  angelic: preorder "op \<prec>" "\<lambda>v v'. v \<prec> v' \<and> \<not> v' \<prec> v" +
  assumes star_distr_Join: "v \<star> \<Squnion>V = \<Squnion>{v''. \<exists>v' \<in> V. v'' = v \<star> v'}"
  and entail_closure: "v \<Turnstile> v' \<Longrightarrow> v \<prec> v'"
  and Join_closure: "(\<forall>v \<in> V. v \<prec> v') \<Longrightarrow> \<Squnion>V \<prec> v'"
  and locality: "v \<prec> v' \<Longrightarrow> v \<star> v'' \<prec> v' \<star> v''"
  and erasure_monotone: "v \<prec> v' \<Longrightarrow> \<lfloor>v\<rfloor> \<subseteq> \<lfloor>v'\<rfloor>"
  and Join_homomorphism: "\<lfloor>\<Squnion>V\<rfloor> = \<Union>{S. \<exists>v \<in> V. S = \<lfloor>v\<rfloor>}"
  and axioms_are_sound: "(v, a, v') \<in> Axioms \<Longrightarrow> a of_states \<lfloor>v \<star> v''\<rfloor> \<subseteq> \<lfloor>v' \<star> v''\<rfloor>"
begin
lemma star_distr_Join': "\<Squnion>V \<star> v = \<Squnion>{v''. \<exists>v' \<in> V. v'' = v' \<star> v}"
  by (auto simp add: views_monoid.commute star_distr_Join)
definition traces_of_view :: "'a \<Rightarrow> 'b description" where
  "traces_of_view v \<equiv> \<Union>{P. \<exists>\<sigma> \<in> \<lfloor>v\<rfloor>. P = ic_traces_ending_in_state(\<sigma>)}"
lemma traces_of_view_and_erasure: "traces_of_view(v) \<subseteq> traces_of_view(v') \<longleftrightarrow> \<lfloor>v\<rfloor> \<subseteq> \<lfloor>v'\<rfloor>"
proof -
  have a: "traces_of_view(v) \<subseteq> traces_of_view(v') \<Longrightarrow> \<lfloor>v\<rfloor> \<subseteq> \<lfloor>v'\<rfloor>"
  proof -
    assume "traces_of_view(v) \<subseteq> traces_of_view(v')"
    then have "\<forall>t. t \<in> \<Union>{P. \<exists>\<sigma> \<in> \<lfloor>v\<rfloor>. P = ic_traces_ending_in_state(\<sigma>)} \<longrightarrow> t \<in> \<Union>{P. \<exists>\<sigma> \<in> \<lfloor>v'\<rfloor>. P = ic_traces_ending_in_state(\<sigma>)}" by (auto simp add: traces_of_view_def Collect_def)
    then have b: "\<forall>t. (\<exists>\<sigma> \<in> \<lfloor>v\<rfloor>. t \<in> ic_traces_ending_in_state(\<sigma>)) \<longrightarrow> (\<exists>\<sigma> \<in> \<lfloor>v'\<rfloor>. t \<in> ic_traces_ending_in_state(\<sigma>))" by auto
    have "\<forall>\<sigma>. \<sigma> \<in> \<lfloor>v\<rfloor> \<longrightarrow> \<sigma> \<in> \<lfloor>v'\<rfloor>"
    proof (rule allI)
      fix \<sigma>
      show "\<sigma> \<in> \<lfloor>v\<rfloor> \<longrightarrow> \<sigma> \<in> \<lfloor>v'\<rfloor>"
      proof (rule impI)
        assume "\<sigma> \<in> \<lfloor>v\<rfloor>"
        then have "\<exists>\<sigma>' \<in> \<lfloor>v\<rfloor>. [(\<sigma>,\<sigma>)] \<in> ic_traces_ending_in_state(\<sigma>')" by (metis trivial_ic_trace_ending_in_state)
        then have "\<exists>\<sigma>' \<in> \<lfloor>v'\<rfloor>. [(\<sigma>,\<sigma>)] \<in> ic_traces_ending_in_state(\<sigma>')" by (metis b)
        then show "\<sigma> \<in> \<lfloor>v'\<rfloor>" by (metis nonempty_mem_ic_traces_ending_in_state1 snd_conv)
      qed
    qed
    then show "\<lfloor>v\<rfloor> \<subseteq> \<lfloor>v'\<rfloor>" by (metis subsetI)
  qed
  have "\<lfloor>v\<rfloor> \<subseteq> \<lfloor>v'\<rfloor> \<Longrightarrow> traces_of_view(v) \<subseteq> traces_of_view(v')" by (auto simp add: traces_of_view_def)
  then show "traces_of_view(v) \<subseteq> traces_of_view(v') \<longleftrightarrow> \<lfloor>v\<rfloor> \<subseteq> \<lfloor>v'\<rfloor>" by (metis a)
qed
lemma traces_of_view_char: "\<forall>\<sigma> \<in> \<lfloor>v\<rfloor>. \<forall>t \<in> ic_traces_ending_in_state(\<sigma>). t \<in> traces_of_view(v)"
  by (auto simp add: traces_of_view_def)
lemma trivial_trace_of_view: "\<forall>\<sigma> \<in> \<lfloor>v\<rfloor>. [(\<sigma>,\<sigma>)] \<in> traces_of_view(v)"
  by (metis traces_of_view_char trivial_ic_trace_ending_in_state)
lemma traces_of_view_Join_homomorphism: "traces_of_view(\<Squnion>V) = \<Union>{P. \<exists>v \<in> V. P = traces_of_view(v)}"
  proof - from traces_of_view_def Join_homomorphism show ?thesis by auto qed
end

locale basic_views_calculus = views
begin
definition b_triple :: "'a \<Rightarrow> 'b description \<Rightarrow> 'a \<Rightarrow> bool" ("_ \<lless>_\<ggreater> _" [54,54,54] 53)
  where "v \<lless>P\<ggreater> v' \<equiv> traces_of_view(v) \<union> Inconsistent \<lbrace>P\<rbrace> traces_of_view(v') \<union> Inconsistent"
lemma b_triple_alt: "v \<lless>P\<ggreater> v' \<longleftrightarrow> (\<forall>tv \<in> traces_of_view(v). \<forall>tp \<in> P. ic(tv@tp) \<longrightarrow> tv@tp \<in> traces_of_view(v'))"
proof -
  have "v \<lless>P\<ggreater> v' \<longleftrightarrow> (traces_of_view(v) \<union> Inconsistent);P \<subseteq> traces_of_view(v') \<union> Inconsistent" by (simp add: b_triple_def hoare_def)
  also have "\<dots> \<longleftrightarrow> (traces_of_view(v);P \<union> Inconsistent;P) \<subseteq> traces_of_view(v') \<union> Inconsistent" by (metis lang_concat_distr_union1)
  also have "\<dots> \<longleftrightarrow> traces_of_view(v);P \<subseteq> traces_of_view(v') \<union> Inconsistent" by (simp, metis Un_commute Un_upper1 inconsistency_persists subset_trans)
  also have "\<dots> \<longleftrightarrow> (\<forall>t. t \<in> traces_of_view(v);P \<longrightarrow> t \<in> traces_of_view(v') \<or> t \<in> Inconsistent)" by auto
  also have "\<dots> \<longleftrightarrow> (\<forall>t. (\<exists>tv tp. tv \<in> traces_of_view(v) \<and> tp \<in> P \<and> t = tv@tp) \<longrightarrow> t \<in> traces_of_view(v') \<or> t \<in> Inconsistent)" by (metis elem_of_lang_concat)
  also have "\<dots> \<longleftrightarrow> (\<forall>tv \<in> traces_of_view(v). \<forall>tp \<in> P. tv@tp \<in> traces_of_view(v') \<or> tv@tp \<in> Inconsistent)" by (auto simp add: Inconsistent_def)
  also have "\<dots> \<longleftrightarrow> (\<forall>tv \<in> traces_of_view(v). \<forall>tp \<in> P. ic(tv@tp) \<longrightarrow> tv@tp \<in> traces_of_view(v'))" by (metis mem_Inconsistent_iff)
  finally show ?thesis .
qed
lemma b_triple_for_atom: "v \<lless>Rep_atom a\<ggreater> v' \<longleftrightarrow> a of_states \<lfloor>v\<rfloor> \<subseteq> \<lfloor>v'\<rfloor>"
proof -
  have b: "v \<lless>Rep_atom a\<ggreater> v' \<Longrightarrow> a of_states \<lfloor>v\<rfloor> \<subseteq> \<lfloor>v'\<rfloor>"
  proof -
    assume "v \<lless>Rep_atom a\<ggreater> v'"
    then have c: "\<forall>tv \<in> traces_of_view(v). \<forall>tp \<in> Rep_atom(a). ic(tv@tp) \<longrightarrow> tv@tp \<in> traces_of_view(v')" by (simp add: b_triple_alt)
    have "\<not> a of_states \<lfloor>v\<rfloor> \<subseteq> \<lfloor>v'\<rfloor> \<Longrightarrow> False"
    proof -
      assume "\<not> a of_states \<lfloor>v\<rfloor> \<subseteq> \<lfloor>v'\<rfloor>"
      then have "\<exists>\<sigma>'. \<sigma>' \<in> a of_states \<lfloor>v\<rfloor> \<and> \<sigma>' \<notin> \<lfloor>v'\<rfloor>" by auto
      then obtain \<sigma>' where "\<sigma>' \<in> a of_states \<lfloor>v\<rfloor> \<and> \<sigma>' \<notin> \<lfloor>v'\<rfloor>" by blast
      then have "\<exists>\<sigma> \<in> \<lfloor>v\<rfloor>. [(\<sigma>,\<sigma>')] \<in> Rep_atom(a) \<and> \<sigma>' \<notin> \<lfloor>v'\<rfloor>" by (auto simp add: atom_applied_to_states_def atom_applied_to_state_def)
      then obtain \<sigma> where d: "\<sigma> \<in> \<lfloor>v\<rfloor>" and e: "[(\<sigma>,\<sigma>')] \<in> Rep_atom(a)" and f: "\<sigma>' \<notin> \<lfloor>v'\<rfloor>" by blast
      have "[(\<sigma>,\<sigma>)] \<in> traces_of_view(v)" by (metis d trivial_trace_of_view)
      then have "\<forall>tp \<in> Rep_atom(a). ic([(\<sigma>,\<sigma>)]@tp) \<longrightarrow> [(\<sigma>,\<sigma>)]@tp \<in> traces_of_view(v')" by (metis c)
      then have "ic([(\<sigma>,\<sigma>)]@[(\<sigma>,\<sigma>')]) \<longrightarrow> [(\<sigma>,\<sigma>)]@[(\<sigma>,\<sigma>')] \<in> traces_of_view(v')" by (metis e)
      then have "[(\<sigma>,\<sigma>),(\<sigma>,\<sigma>')] \<in> traces_of_view(v')" by (metis append_Cons append_Nil fst_conv ic.simps(2) ic.simps(3) snd_conv)
      then have "\<sigma>' \<in> \<lfloor>v'\<rfloor>" by (unfold traces_of_view_def, auto simp add: nonempty_mem_ic_traces_ending_in_state1 nonempty_mem_ic_traces_ending_in_state2)
      then show "False" by (metis f)
    qed
    then show "a of_states \<lfloor>v\<rfloor> \<subseteq> \<lfloor>v'\<rfloor>" by metis
  qed
  have g: "a of_states \<lfloor>v\<rfloor> \<subseteq> \<lfloor>v'\<rfloor> \<Longrightarrow> v \<lless>Rep_atom a\<ggreater> v'"
  proof -
    assume h: "a of_states \<lfloor>v\<rfloor> \<subseteq> \<lfloor>v'\<rfloor>"
    have "\<forall>tv \<in> traces_of_view(v). \<forall>tp \<in> Rep_atom(a). ic(tv@tp) \<longrightarrow> tv@tp \<in> traces_of_view(v')"
    proof (rule ballI)
      fix tv
      assume "tv \<in> traces_of_view(v)"
      then have "\<exists>\<sigma> \<in> \<lfloor>v\<rfloor>. tv \<in> ic_traces_ending_in_state(\<sigma>)" by (auto simp add: traces_of_view_def)
      then obtain \<sigma> where i: "\<sigma> \<in> \<lfloor>v\<rfloor>" and j: "tv \<in> ic_traces_ending_in_state(\<sigma>)" by blast
      show "\<forall>tp \<in> Rep_atom(a). ic(tv@tp) \<longrightarrow> tv@tp \<in> traces_of_view(v')"
      proof (rule ballI)
        fix tp
        assume k: "tp \<in> Rep_atom(a)"
        then have "\<exists>\<sigma>1 \<sigma>2. tp = [(\<sigma>1,\<sigma>2)]" by (metis trace_of_atom_characterization)
        then obtain \<sigma>1 \<sigma>2 where l: "tp = [(\<sigma>1,\<sigma>2)]" by blast
        show "ic(tv@tp) \<longrightarrow> tv@tp \<in> traces_of_view(v')"
        proof (rule impI)
          assume m: "ic(tv@tp)"
          from l m have q: "tv@tp \<in> ic_traces_ending_in_state(\<sigma>2)" by (auto simp add: ic_traces_ending_in_state_def)
          have r: "\<sigma> = \<sigma>1"
          proof -
            from j have s: "tv \<noteq> []" by (auto simp add: ic_traces_ending_in_state_def)
            from l have t: "tp \<noteq> []" by auto
            from s t m have "snd (last tv) = fst (hd tp)" by (rule consistent_append_nonempties)
            then show "\<sigma> = \<sigma>1" by (metis last_state_of_trace_ending_in_state fst_conv hd.simps j l)
          qed
          from h have u: "\<sigma>2 \<in> \<lfloor>v'\<rfloor>"
          proof -
            from h have "\<forall>\<sigma>3. \<sigma>3 \<in> a of_states \<lfloor>v\<rfloor> \<longrightarrow> \<sigma>3 \<in> \<lfloor>v'\<rfloor>" by (metis subsetD)
            then have "\<forall>\<sigma>3. \<sigma>3 \<in> a of_state \<sigma> \<longrightarrow> \<sigma>3 \<in> \<lfloor>v'\<rfloor>" by (metis atom_applied_to_state_states_relationship i subsetD)
            then have "\<forall>\<sigma>3. [(\<sigma>,\<sigma>3)] \<in> Rep_atom(a) \<longrightarrow> \<sigma>3 \<in> \<lfloor>v'\<rfloor>" by (auto simp add: atom_applied_to_state_def)
            then show "\<sigma>2 \<in> \<lfloor>v'\<rfloor>" by (metis r l k)
          qed
          show "tv@tp \<in> traces_of_view(v')" by (metis u q traces_of_view_char)
        qed
      qed
    qed
    then show "v \<lless>Rep_atom a\<ggreater> v'" by (simp add: b_triple_alt h)
  qed
  from b g show "v \<lless>Rep_atom a\<ggreater> v' \<longleftrightarrow> a of_states \<lfloor>v\<rfloor> \<subseteq> \<lfloor>v'\<rfloor>" ..
qed
lemma Batom: "(v, a, v') \<in> Axioms \<Longrightarrow> v \<lless>Rep_atom a\<ggreater> v'"
  by (metis axioms_are_sound b_triple_for_atom comm_neutral)
lemma Bskip: "v \<lless>skip\<ggreater> v"
  by (metis Hskip b_triple_def)
lemma Bseq: "v \<lless>P\<ggreater> v' \<Longrightarrow> v' \<lless>P'\<ggreater> v'' \<Longrightarrow> v \<lless>P;P'\<ggreater> v''"
  by (unfold b_triple_def, metis Hseq)
lemma Bchoice: "(\<forall>P\<in>X.  v \<lless>P\<ggreater> v') \<Longrightarrow> v \<lless>\<Union>X\<ggreater> v'"
  by (metis Hchoice b_triple_def)
lemma Biter: "v \<lless>P\<ggreater> v \<Longrightarrow> v \<lless>P\<^sup>\<star>\<ggreater> v"
  by (metis Hiter b_triple_def)
lemma Bcons: "v \<prec> v' \<Longrightarrow> v' \<lless>P\<ggreater> v'' \<Longrightarrow> v'' \<prec> v''' \<Longrightarrow> v \<lless>P\<ggreater> v'''"
proof -
  assume a3: "v \<prec> v'" and a4: "v' \<lless>P\<ggreater> v''" and a5: "v'' \<prec> v'''"
  from a3 erasure_monotone have "traces_of_view(v) \<subseteq> traces_of_view(v')" by (metis traces_of_view_and_erasure)
  hence b1: "traces_of_view(v) \<union> Inconsistent \<subseteq> traces_of_view(v') \<union> Inconsistent" by (metis Un_mono subset_refl)
  from a5 erasure_monotone have "traces_of_view(v'') \<subseteq> traces_of_view(v''')" by (metis traces_of_view_and_erasure)
  hence b2: "traces_of_view(v'') \<union> Inconsistent \<subseteq> traces_of_view(v''') \<union> Inconsistent" by (metis Un_mono subset_refl)
  from a4 have b3: "traces_of_view(v') \<union> Inconsistent \<lbrace>P\<rbrace> traces_of_view(v'') \<union> Inconsistent" by (simp add: b_triple_def)
  from b1 b3 b2 have "traces_of_view(v) \<union> Inconsistent \<lbrace>P\<rbrace> traces_of_view(v''') \<union> Inconsistent" by (rule Hcons)
  thus "v \<lless>P\<ggreater> v'''" by (metis b_triple_def)
qed
lemma Bdisj: "(\<forall>v \<in> V. v \<lless>P\<ggreater> v') \<Longrightarrow> \<Squnion>V \<lless>P\<ggreater> v'"
proof -
  assume "\<forall>v \<in> V. v \<lless>P\<ggreater> v'"
  hence "\<forall>v \<in> V. traces_of_view(v) \<union> Inconsistent \<lbrace>P\<rbrace> traces_of_view(v') \<union> Inconsistent" by (metis b_triple_def)
  hence "\<forall>v \<in> V. traces_of_view(v) \<lbrace>P\<rbrace> traces_of_view(v') \<union> Inconsistent"
  proof -
    assume p: "\<forall>v \<in> V. traces_of_view(v) \<union> Inconsistent \<lbrace>P\<rbrace> traces_of_view(v') \<union> Inconsistent"
    show "\<forall>v \<in> V. traces_of_view(v) \<lbrace>P\<rbrace> traces_of_view(v') \<union> Inconsistent"
    proof (rule ballI)
      fix v
      assume "v \<in> V"
      with p have p1: "traces_of_view(v) \<union> Inconsistent \<lbrace>P\<rbrace> traces_of_view(v') \<union> Inconsistent" by auto
      have p2: "traces_of_view(v) \<subseteq> traces_of_view(v) \<union> Inconsistent" by auto
      have p3: "traces_of_view(v') \<union> Inconsistent \<subseteq> traces_of_view(v') \<union> Inconsistent" by simp
      from p2 p1 p3 show "traces_of_view(v) \<lbrace>P\<rbrace> traces_of_view(v') \<union> Inconsistent" by (rule Hcons)
    qed
  qed
  hence "\<forall>Q \<in> {Q. \<exists>v \<in> V. Q = traces_of_view(v)}. Q \<lbrace>P\<rbrace> traces_of_view(v') \<union> Inconsistent" by auto
  hence "\<Union>{Q. \<exists>v \<in> V. Q = traces_of_view(v)} \<lbrace>P\<rbrace> traces_of_view(v') \<union> Inconsistent" by (rule Hdisj)
  hence b1: "traces_of_view(\<Squnion>V) \<lbrace>P\<rbrace> traces_of_view(v') \<union> Inconsistent" by (metis traces_of_view_Join_homomorphism)
  from inconsistency_persists hoare_def have b2: "Inconsistent \<lbrace>P\<rbrace> traces_of_view(v') \<union> Inconsistent" by (metis semilattice_sup_class.le_supI2)
  from b1 b2 have "\<forall>Q \<in> {traces_of_view(\<Squnion>V), Inconsistent}. Q \<lbrace>P\<rbrace> traces_of_view(v') \<union> Inconsistent" by auto
  hence "\<Union>{traces_of_view(\<Squnion>V), Inconsistent} \<lbrace>P\<rbrace> traces_of_view(v') \<union> Inconsistent" by (rule Hdisj)
  thus "\<Squnion>V \<lless>P\<ggreater> v'" by (unfold b_triple_def, auto)
qed
end

locale framing_calculus = basic_views_calculus
begin
definition f_triple :: "'a \<Rightarrow> 'b description \<Rightarrow> 'a \<Rightarrow> bool" ("_ \<langle>_\<rangle> _" [54,54,54] 53)
  where "v \<langle>P\<rangle> v' \<equiv> (\<forall>v''. v \<star> v'' \<lless>P\<ggreater> v' \<star> v'')"
lemma f_triple_b_triple_relationship: "v \<langle>P\<rangle> v' \<Longrightarrow> v \<lless>P\<ggreater> v'"
  by (unfold f_triple_def, metis comm_neutral)
lemma Fatom: "(v, a, v') \<in> Axioms \<Longrightarrow> v \<langle>Rep_atom a\<rangle> v'"
  by (metis axioms_are_sound b_triple_for_atom f_triple_def)
lemma Fskip: "v \<langle>skip\<rangle> v"
  by (metis Bskip f_triple_def)
lemma Fseq: "v \<langle>P\<rangle> v' \<Longrightarrow> v' \<langle>P'\<rangle> v'' \<Longrightarrow> v \<langle>P;P'\<rangle> v''"
  by (unfold f_triple_def, metis Bseq)
lemma Fchoice: "(\<forall>P\<in>X.  v \<langle>P\<rangle> v') \<Longrightarrow> v \<langle>\<Union>X\<rangle> v'"
  by (metis Bchoice f_triple_def)
lemma Fiter: "v \<langle>P\<rangle> v \<Longrightarrow> v \<langle>P\<^sup>\<star>\<rangle> v"
  by (metis Biter f_triple_def)
lemma Fcons: "v \<prec> v' \<Longrightarrow> v' \<langle>P\<rangle> v'' \<Longrightarrow> v'' \<prec> v''' \<Longrightarrow> v \<langle>P\<rangle> v'''"
  by (unfold f_triple_def, metis Bcons locality)
lemma Fdisj: "(\<forall>v \<in> V. v \<langle>P\<rangle> v') \<Longrightarrow> \<Squnion>V \<langle>P\<rangle> v'"
proof -
  assume a: "\<forall>v \<in> V. v \<langle>P\<rangle> v'"
  have "\<forall>v1. \<Squnion>V \<star> v1 \<lless>P\<ggreater> v' \<star> v1"
  proof (rule allI)
    fix v1
    from a have "\<forall>v \<in> V. v \<star> v1 \<lless>P\<ggreater> v' \<star> v1" by (metis f_triple_def)
    hence "\<forall>v \<in> {v2. \<exists>v \<in> V. v2 = v \<star> v1}. v \<lless>P\<ggreater> v' \<star> v1" by auto
    hence "\<Squnion>{v2. \<exists>v \<in> V. v2 = v \<star> v1} \<lless>P\<ggreater> v' \<star> v1" by (rule Bdisj)
    thus "\<Squnion>V \<star> v1 \<lless>P\<ggreater> v' \<star> v1" by (metis star_distr_Join')
  qed
  thus "\<Squnion>V \<langle>P\<rangle> v'" by (metis f_triple_def)
qed
lemma Fframe: "v \<langle>P\<rangle> v' \<Longrightarrow> v \<star> v'' \<langle>P\<rangle> v' \<star> v''"
  by (metis assoc f_triple_def)
end

locale full_views_calculus = framing_calculus
begin
fun as_triple :: "'a \<Rightarrow> 'b atom_sequence \<Rightarrow> 'a \<Rightarrow> bool" ("_ ~_~ _" [54,54,54] 53) where 
    "v ~[]~ v'    =  v \<langle>skip\<rangle> v'"
  | "v ~a#as~ v'  =  (\<exists>v''. v \<langle>Rep_atom a\<rangle> v'' \<and> v''~as~ v')"
definition views_triple :: "'a \<Rightarrow> 'b command \<Rightarrow> 'a \<Rightarrow> bool" ("\<lbrace>_\<rbrace> _ \<lbrace>_\<rbrace>" [54,54,54] 53)
  where "\<lbrace>v\<rbrace> C \<lbrace>v'\<rbrace> \<equiv> (\<forall>as \<in> C. v ~as~ v')"
lemma as_triple_f_triple_relationship: "\<And>v v'. v ~as~ v' \<Longrightarrow> v \<langle>trace_set_of_atom_sequence(as)\<rangle> v'"
  by (induct as, auto simp add: Fseq)
theorem views_triple_f_triple_relationship: "\<lbrace>v\<rbrace> C \<lbrace>v'\<rbrace> \<Longrightarrow> v \<langle>trace_set_of_command(C)\<rangle> v'"
proof (unfold trace_set_of_command_def views_triple_def)
  assume "\<forall>as\<in>C. v ~as~ v'"
  hence "\<forall>as\<in>C. v \<langle>trace_set_of_atom_sequence(as)\<rangle> v'" by (metis as_triple_f_triple_relationship)
  hence "\<forall>P \<in> {P. \<exists>as\<in>C. P = trace_set_of_atom_sequence(as)}. v \<langle>P\<rangle> v'" by auto
  with Fchoice show "v \<langle>\<Union>{P. \<exists>as\<in>C. P = trace_set_of_atom_sequence(as)}\<rangle> v'" by auto
qed
lemma Vatom: "(v, a, v') \<in> Axioms \<Longrightarrow> \<lbrace>v\<rbrace> {[a]} \<lbrace>v'\<rbrace>"
proof -
  assume a: "(v, a, v') \<in> Axioms"
  from Fskip have c: "v' ~[]~ v'" by (metis as_triple.simps(1))
  from a Fatom have d: "v \<langle>Rep_atom a\<rangle> v'" by metis
  from c d have "v ~[a]~ v'" by (metis as_triple.simps(2))
  thus "\<lbrace>v\<rbrace> {[a]} \<lbrace>v'\<rbrace>" by (metis emptyE insertE views_triple_def)
qed
lemma Vskip: "\<lbrace>v\<rbrace> skip \<lbrace>v\<rbrace>"
  by (metis Fskip as_triple.simps(1) elem_of_skip views_triple_def)
lemma as_triple_seq: "\<And>v v' as' v''. v ~as~ v' \<Longrightarrow> v' ~as'~ v'' \<Longrightarrow> v ~as@as'~ v''"
proof -
  have helper: "\<And>as' v v' v''. v \<langle>skip\<rangle> v' \<Longrightarrow> v' ~as'~ v'' \<Longrightarrow> v ~as'~ v''"
  proof - fix as' show "\<And>v v' v''. v \<langle>skip\<rangle> v' \<Longrightarrow> v' ~as'~ v'' \<Longrightarrow> v ~as'~ v''"
    by (cases as', simp, metis Fseq lang_concat_unit1, simp, metis Fseq lang_concat_unit1) 
  qed
  show "\<And>v v' as' v''. v ~as~ v' \<Longrightarrow> v' ~as'~ v'' \<Longrightarrow> v ~as@as'~ v''"
    by (induct as, simp, metis helper, auto)
qed
lemma Vseq: "\<lbrace>v\<rbrace> C \<lbrace>v'\<rbrace> \<Longrightarrow> \<lbrace>v'\<rbrace> C' \<lbrace>v''\<rbrace> \<Longrightarrow> \<lbrace>v\<rbrace> C;C' \<lbrace>v''\<rbrace>"
proof (unfold views_triple_def)
  assume a: "\<forall>as \<in> C. v ~as~ v'"
  assume b: "\<forall>as' \<in> C'. v' ~as'~ v''"
  show "\<forall>as \<in> (C;C'). v ~as~ v''"
  proof (rule ballI)
    fix as
    assume "as \<in> (C;C')"
    then have "\<exists>asc asc'. asc \<in> C \<and> asc' \<in> C' \<and> as = asc@asc'" by (metis elem_of_lang_concat)
    then obtain asc asc' where c1: "asc \<in> C" and c2: "asc' \<in> C'" and c3: "as = asc@asc'" by blast
    from c1 a have d1: "v ~asc~ v'" by auto
    from c2 b have d2: "v' ~asc'~ v''" by auto
    from d1 d2 c3 show "v ~as~ v''" by (metis as_triple_seq)
  qed
qed
lemma Vchoice: "(\<forall>C \<in> Y. \<lbrace>v\<rbrace> C \<lbrace>v'\<rbrace>) \<Longrightarrow> \<lbrace>v\<rbrace> \<Union>Y \<lbrace>v'\<rbrace>"
  by (metis UnionE views_triple_def)
lemma Viterhelper: "\<lbrace>v\<rbrace> C \<lbrace>v\<rbrace> \<Longrightarrow> \<lbrace>v\<rbrace> iter C n \<lbrace>v\<rbrace>"
  by (induct n, simp, metis Vskip, auto simp add: Vseq)
lemma Viter: "\<lbrace>v\<rbrace> C \<lbrace>v\<rbrace> \<Longrightarrow> \<lbrace>v\<rbrace> C\<^sup>\<star> \<lbrace>v\<rbrace>"
proof -
  assume "\<lbrace>v\<rbrace> C \<lbrace>v\<rbrace>"
  with Viterhelper have "\<forall>C' \<in> {C'. \<exists>n. C' = iter C n}. \<lbrace>v\<rbrace> C' \<lbrace>v\<rbrace>" by auto
  hence "\<lbrace>v\<rbrace> \<Union>{C'. \<exists>n. C' = iter C n} \<lbrace>v\<rbrace>" by (rule Vchoice)
  thus "\<lbrace>v\<rbrace> C\<^sup>\<star> \<lbrace>v\<rbrace>" by (metis kleene_star_def)
qed
lemma as_triple_cons: "\<And>v v' v'' v'''. v \<prec> v' \<Longrightarrow> v' ~as~ v'' \<Longrightarrow> v'' \<prec> v''' \<Longrightarrow> v ~as~ v'''"
  by (induct as, simp, metis Fcons, simp, metis Fcons angelic.order_refl)
lemma Vcons: "v \<prec> v' \<Longrightarrow> \<lbrace>v'\<rbrace> C \<lbrace>v''\<rbrace> \<Longrightarrow> v'' \<prec> v''' \<Longrightarrow> \<lbrace>v\<rbrace> C \<lbrace>v'''\<rbrace>"
  by (unfold views_triple_def, metis as_triple_cons)
lemma as_triple_disj: "\<And>v' V. (\<forall>v \<in> V. v ~as~ v') \<Longrightarrow> \<Squnion>V ~as~ v'"
proof (induct as, simp, metis Fdisj, simp)
  fix a as' v' V
  assume IH: "\<And>v' V. \<forall>v\<in>V. v ~as'~ v' \<Longrightarrow> \<Squnion>V ~as'~ v'"
  assume a4: "\<forall>v\<in>V. \<exists>v''. v \<langle>a\<rangle> v'' \<and> v'' ~as'~ v'"
  let ?Vq = "{v1. \<exists>v \<in> V. v \<langle>a\<rangle> v1 \<and> v1 ~as'~ v'}"
  let ?v2 = "\<Squnion>?Vq"
  have b2: "\<Squnion>V \<langle>a\<rangle> ?v2"
  proof -
    have c: "\<forall>v \<in> V. v \<langle>a\<rangle> ?v2"
    proof (rule ballI)
      fix v
      assume c1: "v \<in> V"
      with a4 have "\<exists>v''. v \<langle>a\<rangle> v'' \<and> v'' ~as'~ v'" by auto
      then obtain v'' where c2: "v \<langle>a\<rangle> v''" and c3: "v'' ~as'~ v'" by blast
      from c1 c2 c3 have "v'' \<in> ?Vq" by auto
      hence "v'' \<Turnstile> ?v2" by (auto simp add: Collect_def Sup_upper)
      hence c4: "v'' \<prec> ?v2" by (simp add: entail_closure)
      have c5: "v \<prec> v" by (metis angelic.order_refl)
      from c5 c2 c4 show "v \<langle>a\<rangle> ?v2" by (rule Fcons)
    qed
    thus ?thesis by (rule Fdisj)
  qed
  have b3: "?v2 ~as'~ v'" by (auto simp add: IH)
  from b2 b3 show "\<exists>v2. \<Squnion>V \<langle>a\<rangle> v2 \<and> v2 ~as'~ v'" by auto
qed
lemma Vdisj: "(\<forall>v \<in> V. \<lbrace>v\<rbrace> C \<lbrace>v'\<rbrace>) \<Longrightarrow> \<lbrace>\<Squnion>V\<rbrace> C \<lbrace>v'\<rbrace>"
  by (unfold views_triple_def, metis as_triple_disj)
lemma as_triple_frame: "\<And>v v' v''. v ~as~ v' \<Longrightarrow> v \<star> v'' ~as~ v' \<star> v''"
proof (induct as, simp, metis Fframe)
  fix a as v v' v''
  assume IH: "\<And>v v' v''. v ~as~ v' \<Longrightarrow> v \<star> v'' ~as~ v' \<star> v''"
  assume "v ~a#as~ v'"
  hence "\<exists>vp. v \<langle>Rep_atom a\<rangle> vp \<and> vp~as~ v'" by simp
  then obtain vp where a3: "v \<langle>Rep_atom a\<rangle> vp" and a4: "vp ~as~ v'" by blast
  from a3 have a6: "v \<star> v'' \<langle>Rep_atom a\<rangle> vp \<star> v''" by (rule Fframe)
  from a4 have a7: "vp \<star> v'' ~as~ v' \<star> v''" by (rule IH)
  from a6 a7 show "v \<star> v'' ~a#as~ v' \<star> v''" by (auto simp add: f_triple_def)
qed
lemma Vframe: "\<lbrace>v\<rbrace> C \<lbrace>v'\<rbrace> \<Longrightarrow> \<lbrace>v\<star>v''\<rbrace> C \<lbrace>v'\<star>v''\<rbrace>"
  by (metis as_triple_frame views_triple_def)
lemma as_triple_conc: "\<And>v1 as1 v1' v2 as2 v2'. v1 ~as1~ v1' \<Longrightarrow> v2 ~as2~ v2' \<Longrightarrow> as \<in> as1\<otimes>as2 \<Longrightarrow> v1 \<star> v2 ~as~ v1' \<star> v2'"
proof (induct as)
  show "\<And>v1 as1 v1' v2 as2 v2'.
       v1 ~as1~ v1' \<Longrightarrow> v2 ~as2~ v2' \<Longrightarrow> [] \<in> as1 \<otimes> as2 \<Longrightarrow> v1 \<star> v2 ~[]~ v1' \<star> v2'"
  proof - fix v1 as1 v1' v2 as2 v2'
    assume a1: "v1 ~as1~ v1'"
    assume a2: "v2 ~as2~ v2'"
    assume "[] \<in> as1 \<otimes> as2"
    hence a3: "as1 = [] \<and> as2 = []" by (metis interleave_contains_empty)
    from a1 a3 have "v1 \<langle>skip\<rangle> v1'" by simp
    hence a5: "v1 \<star> v2 \<langle>skip\<rangle> v1' \<star> v2" by (rule Fframe)
    from a2 a3 have "v2 \<langle>skip\<rangle> v2'" by simp
    with Fframe have a6: "v1' \<star> v2 \<langle>skip\<rangle> v1' \<star> v2'" by (metis commute)
    from a5 a6 show "v1 \<star> v2 ~[]~ v1' \<star> v2'" by (simp, metis Fseq lang_concat_unit1)
  qed
next
  show "\<And>a as v1 as1 v1' v2 as2 v2'.
       (\<And>v1 as1 v1' v2 as2 v2'. v1 ~as1~ v1' \<Longrightarrow> v2 ~as2~ v2' \<Longrightarrow> as \<in> as1\<otimes>as2 \<Longrightarrow> v1 \<star> v2 ~as~ v1' \<star> v2') \<Longrightarrow>
       v1 ~as1~ v1' \<Longrightarrow> v2 ~as2~ v2' \<Longrightarrow> a#as \<in> as1\<otimes>as2 \<Longrightarrow> v1 \<star> v2 ~(a#as)~ v1' \<star> v2'"
  proof - fix a as v1 as1 v1' v2 as2 v2'
    assume IH: "(\<And>v1 as1 v1' v2 as2 v2'. v1 ~as1~ v1' \<Longrightarrow> v2 ~as2~ v2' \<Longrightarrow> as \<in> as1 \<otimes> as2 \<Longrightarrow> v1 \<star> v2 ~as~ v1' \<star> v2')"
    assume b1: "v1 ~as1~ v1'" and b2: "v2 ~as2~ v2'" and b3: "a#as \<in> as1\<otimes>as2"
    from b3 have "(\<exists>t. as1 = a#t \<and> as \<in> t \<otimes> as2) \<or> (\<exists>t'. as2 = a#t' \<and> as \<in> as1 \<otimes> t')" by (simp add: interleave_contains_nonempty)
    moreover {
      assume "\<exists>t. as1 = a#t \<and> as \<in> t \<otimes> as2"
      then obtain t where c1: "as1 = a#t" and c2: "as \<in> t \<otimes> as2" by blast
      from c1 b1 have "\<exists>v1p. v1 \<langle>Rep_atom a\<rangle> v1p \<and> v1p ~t~ v1'" by simp
      then obtain v1p where c3: "v1 \<langle>Rep_atom a\<rangle> v1p" and c4: "v1p ~t~ v1'" by blast
      from c4 b2 c2 have c6: "v1p \<star> v2 ~as~ v1' \<star> v2'" by (rule IH)
      from c3 have "v1 \<star> v2 \<langle>Rep_atom a\<rangle> v1p \<star> v2" by (rule Fframe)
      with c6 have "v1 \<star> v2 ~(a#as)~ v1' \<star> v2'" by (metis as_triple.simps(2))
    } moreover {
      assume "\<exists>t'. as2 = a#t' \<and> as \<in> as1 \<otimes> t'"
      then obtain t' where c1: "as2 = a#t'" and c2: "as \<in> as1 \<otimes> t'" by blast
      from c1 b2 have "\<exists>v2p. v2 \<langle>Rep_atom a\<rangle> v2p \<and> v2p ~t'~ v2'" by simp
      then obtain v2p where c3: "v2 \<langle>Rep_atom a\<rangle> v2p" and c4: "v2p ~t'~ v2'" by blast
      from b1 c4 c2 have c6: "v1 \<star> v2p ~as~ v1' \<star> v2'" by (rule IH)
      from c3 have "v1 \<star> v2 \<langle>Rep_atom a\<rangle> v1 \<star> v2p" by (metis Fframe commute)
      with c6 have "v1 \<star> v2 ~(a#as)~ v1' \<star> v2'" by (metis as_triple.simps(2))
    }
    ultimately show "v1 \<star> v2 ~(a#as)~ v1' \<star> v2'" by (rule disjE)
  qed
qed
lemma Vconc: "\<lbrace>v1\<rbrace> C1 \<lbrace>v1'\<rbrace> \<Longrightarrow> \<lbrace>v2\<rbrace> C2 \<lbrace>v2'\<rbrace> \<Longrightarrow> \<lbrace>v1\<star>v2\<rbrace> C1\<parallel>C2 \<lbrace>v1'\<star>v2'\<rbrace>"
proof (unfold views_triple_def)
  assume a: "\<forall>as1 \<in> C1. v1 ~as1~ v1'"
  assume b: "\<forall>as2 \<in> C2. v2 ~as2~ v2'"
  show "\<forall>as \<in> (C1\<parallel>C2). v1 \<star> v2 ~as~ v1' \<star> v2'"
  proof (rule ballI)
    fix as
    assume "as \<in> (C1\<parallel>C2)"
    then have "\<exists>as1 as2. as1 \<in> C1 \<and> as2 \<in> C2 \<and> as \<in> as1\<otimes>as2" by (metis elem_of_lang_interleave)
    then obtain as1 as2 where c1: "as1 \<in> C1" and c2: "as2 \<in> C2" and c3: "as \<in> as1\<otimes>as2" by blast
    from c1 a have d1: "v1 ~as1~ v1'" by auto
    from c2 b have d2: "v2 ~as2~ v2'" by auto
    from d1 d2 c3 show "v1 \<star> v2 ~as~ v1' \<star> v2'" by (metis as_triple_conc)
  qed
qed
end

end
