(*  Title:       DescriptionsAndCommands.thy
    Author:      Stephan van Staden, ETH Zurich, Switzerland
    Description: States, traces, descriptions, atoms and commands.
    Copyright    2012 ETH Zurich
*)

theory DescriptionsAndCommands
  imports Languages
begin

type_synonym 'a pair = "('a \<times> 'a)"

type_synonym 'a trace = "'a pair list"

type_synonym 'a description = "'a pair language"

typedef 'a atom = "{P::'a description. \<forall>t \<in> P. length t = 1}"
  by auto

definition atom_applied_to_state :: "'a atom \<Rightarrow> 'a \<Rightarrow> 'a set" (infix "of'_state" 53) where
  "a of_state \<sigma> \<equiv> {\<sigma>'. [(\<sigma>, \<sigma>')] \<in> Rep_atom a}"
definition atom_applied_to_states :: "'a atom \<Rightarrow> 'a set \<Rightarrow> 'a set" (infix "of'_states" 53) where
  "a of_states S \<equiv> \<Union>{S'. \<exists>\<sigma> \<in> S. S' = a of_state \<sigma>}"
fun ic :: "'a trace \<Rightarrow> bool" where
   "ic [] = True"
 | "ic [_] = True"
 | "ic (e1#e2#t) = (snd e1 = fst e2 \<and> ic (e2#t))"
definition ic_traces_ending_in_state :: "'a \<Rightarrow> 'a description" where
  "ic_traces_ending_in_state \<sigma> \<equiv> {t. ic(t) \<and> (\<exists>t' \<sigma>'. t = t'@[(\<sigma>',\<sigma>)])}"
lemma trace_of_atom_characterization: "\<forall>t \<in> Rep_atom a. \<exists>\<sigma> \<sigma>'. t = [(\<sigma>,\<sigma>')]"
proof -
  have "\<forall>P \<in> atom. \<forall>t \<in> P. length t = 1" by (auto simp add: atom_def)
  thus ?thesis by (metis Rep_atom butlast.simps(2) butlast_conv_take length_butlast neq_Nil_conv order_eq_iff prod.exhaust take_1_Cons take_all)
qed
lemma all_traces_of_atom_ic: "\<forall>t \<in> Rep_atom a. ic(t)"
  by (metis ic.simps(2) trace_of_atom_characterization)
lemma atom_applied_to_states_monotone: "S \<subseteq> S' \<Longrightarrow> a of_states S \<subseteq> a of_states S'"
  by (auto simp add: atom_applied_to_states_def)
lemma atom_applied_to_state_states_relationship: "\<sigma> \<in> S \<Longrightarrow> a of_state \<sigma> \<subseteq> a of_states S"
  by (auto simp add: atom_applied_to_states_def)
lemma ic_cons_iff: "ic(e#t) \<longleftrightarrow> ic(t) \<and> (t = [] \<or> snd e = fst (hd t))"
  by (induct t, auto)
lemma consistent_append: "ic(t@t') \<Longrightarrow> ic(t) \<and> ic(t')"
proof -
  have helper: "\<And>t'. ic(t@t') \<Longrightarrow> ic(t) \<and> ic(t')"
  proof (induct t, simp)
    fix a t t2
    assume IH: "\<And>t'. ic (t @ t') \<Longrightarrow> ic t \<and> ic t'"
    assume "ic ((a # t) @ t2)"
    then have "ic (t @ t2) \<and> (t@t2 = [] \<or> snd a = fst (hd (t@t2)))" by (simp add: ic_cons_iff)
    then have "ic t \<and> ic t2 \<and> (t@t2 = [] \<or> snd a = fst (hd (t@t2)))" by (metis IH)
    then show "ic (a # t) \<and> ic t2" by (metis Nil_is_append_conv hd_append ic_cons_iff)
  qed
  assume "ic(t@t')"
  then show "ic(t) \<and> ic(t')" by (rule helper)
qed
lemma consistent_append_nonempties: "t \<noteq> [] \<Longrightarrow> t' \<noteq> [] \<Longrightarrow> ic(t@t') \<Longrightarrow> snd (last t) = fst (hd t')"
  by (metis Cons_eq_appendI append_eq_appendI consistent_append eq_Nil_appendI ic_cons_iff last_snoc rev_cases)
lemma trivial_ic_trace_ending_in_state: "[(\<sigma>1,\<sigma>)] \<in> ic_traces_ending_in_state(\<sigma>)"
  by (simp add: ic_traces_ending_in_state_def)
lemma trivial_ic_trace_ending_in_state_length2: "[(\<sigma>1,\<sigma>2),(\<sigma>2,\<sigma>)] \<in> ic_traces_ending_in_state(\<sigma>)"
  by (auto simp add: ic_traces_ending_in_state_def)
lemma ic_traces_ending_in_state_not_empty: "ic_traces_ending_in_state(\<sigma>) \<noteq> {}"
  by (metis trivial_ic_trace_ending_in_state emptyE)
lemma trace_can_end_only_in_one_state: "t \<in> ic_traces_ending_in_state(\<sigma>) \<Longrightarrow> t \<in> ic_traces_ending_in_state(\<sigma>') \<Longrightarrow> \<sigma> = \<sigma>'"
  by (unfold ic_traces_ending_in_state_def, auto)
lemma last_state_of_trace_ending_in_state: "t \<in> ic_traces_ending_in_state(\<sigma>) \<Longrightarrow> snd (last t) = \<sigma>"
  by (auto simp add: ic_traces_ending_in_state_def)
lemma nonempty_mem_ic_traces_ending_in_state1: "t = [] \<Longrightarrow> a#t \<in> ic_traces_ending_in_state \<sigma> \<longleftrightarrow> snd a = \<sigma>"
proof -
  assume a: "t = []"
  show "a#t \<in> ic_traces_ending_in_state \<sigma> \<longleftrightarrow> snd a = \<sigma>"
  proof (rule iffI)
    assume "a#t \<in> ic_traces_ending_in_state \<sigma>"
    thus "snd a = \<sigma>" by (metis a append_butlast_last_id butlast_snoc hd.simps last_state_of_trace_ending_in_state not_Cons_self self_append_conv2)
  next
    assume b: "snd a = \<sigma>"
    hence "\<exists>\<sigma>1. a = (\<sigma>1,\<sigma>)" by (metis pair_collapse)
    with a have "\<exists>\<sigma>1. a#t = [(\<sigma>1,\<sigma>)]" by auto
    thus "a#t \<in> ic_traces_ending_in_state \<sigma>" by (metis trivial_ic_trace_ending_in_state)
  qed
qed
lemma nonempty_mem_ic_traces_ending_in_state2: "t \<noteq> [] \<Longrightarrow> a#t \<in> ic_traces_ending_in_state \<sigma> \<longleftrightarrow> snd a = fst (hd t) \<and> t \<in> ic_traces_ending_in_state \<sigma>"
proof -
  assume a: "t \<noteq> []"
  then have helper: "\<And>b. (\<exists>t'. a#t = t'@[b]) \<longleftrightarrow> (\<exists>t'. t = t'@[b])" by (metis append_Cons append_butlast_last_id last_snoc)
  have b1: "a#t \<in> ic_traces_ending_in_state \<sigma> \<longleftrightarrow> ic(a#t) \<and> (\<exists>t' \<sigma>'. a#t = t'@[(\<sigma>',\<sigma>)])" by (simp add: ic_traces_ending_in_state_def)
  have b2: "\<dots> \<longleftrightarrow> ic(t) \<and> (t = [] \<or> snd a = fst (hd t)) \<and> (\<exists>t' \<sigma>'. a#t = t'@[(\<sigma>',\<sigma>)])" by (simp add: ic_cons_iff)
  have b3: "\<dots> \<longleftrightarrow> ic(t) \<and> snd a = fst (hd t) \<and> (\<exists>t' \<sigma>'. a#t = t'@[(\<sigma>',\<sigma>)])" by (metis a)
  have b4: "\<dots> \<longleftrightarrow> snd a = fst (hd t) \<and> ic(t) \<and> (\<exists>t' \<sigma>'. t = t'@[(\<sigma>',\<sigma>)])" by (metis helper)
  have b5: "\<dots> \<longleftrightarrow> snd a = fst (hd t) \<and> t \<in> ic_traces_ending_in_state \<sigma>" by (simp add: ic_traces_ending_in_state_def)
  from b1 b2 b3 b4 b5 show "a#t \<in> ic_traces_ending_in_state \<sigma> \<longleftrightarrow> snd a = fst (hd t) \<and> t \<in> ic_traces_ending_in_state \<sigma>" by metis
qed
lemma ic_traces_ending_in_state_snoc: "t@[a] \<in> ic_traces_ending_in_state(\<sigma>) \<longleftrightarrow> (\<exists>\<sigma>'. a = (\<sigma>',\<sigma>) \<and> (t = [] \<or> t \<in> ic_traces_ending_in_state(\<sigma>')))"
proof (induct t, simp)
  show "([a] \<in> ic_traces_ending_in_state \<sigma>) \<longleftrightarrow> (\<exists>\<sigma>'. a = (\<sigma>', \<sigma>))" by (auto simp add: ic_traces_ending_in_state_def)
next
  fix aa t
  assume IH: "(t @ [a] \<in> ic_traces_ending_in_state \<sigma>) \<longleftrightarrow> (\<exists>\<sigma>'. a = (\<sigma>', \<sigma>) \<and> (t = [] \<or> t \<in> ic_traces_ending_in_state \<sigma>'))"
  show "((aa # t) @ [a] \<in> ic_traces_ending_in_state \<sigma>) \<longleftrightarrow> (\<exists>\<sigma>'. a = (\<sigma>', \<sigma>) \<and> (aa # t = [] \<or> aa # t \<in> ic_traces_ending_in_state \<sigma>'))"
  proof -
    have a: "t @ [a] \<noteq> []" by (metis snoc_eq_iff_butlast)
    have "(aa # t) @ [a] \<in> ic_traces_ending_in_state \<sigma> \<longleftrightarrow> aa # (t @ [a]) \<in> ic_traces_ending_in_state \<sigma>" by (metis append_Cons)
    also from a have "\<dots> \<longleftrightarrow> snd aa = fst (hd (t @ [a])) \<and> (t @ [a]) \<in> ic_traces_ending_in_state \<sigma>" by (metis nonempty_mem_ic_traces_ending_in_state2)
    also have "\<dots> \<longleftrightarrow> snd aa = fst (hd (t @ [a])) \<and> (\<exists>\<sigma>'. a = (\<sigma>', \<sigma>) \<and> (t = [] \<or> t \<in> ic_traces_ending_in_state \<sigma>'))" by (metis IH)
    also have "\<dots> \<longleftrightarrow> (\<exists>\<sigma>'. a = (\<sigma>', \<sigma>) \<and> (aa # t = [] \<or> aa # t \<in> ic_traces_ending_in_state \<sigma>'))"
      by (cases t, auto simp add: ic_traces_ending_in_state_def, metis pair_collapse, metis append_butlast_last_id last.simps last_snoc list.simps(3))
    finally show "((aa # t) @ [a] \<in> ic_traces_ending_in_state \<sigma>) \<longleftrightarrow> (\<exists>\<sigma>'. a = (\<sigma>', \<sigma>) \<and> (aa # t = [] \<or> aa # t \<in> ic_traces_ending_in_state \<sigma>'))" by auto
  qed
qed
lemma history_independence: "t1 \<in> ic_traces_ending_in_state(\<sigma>) \<Longrightarrow> t2 \<in> ic_traces_ending_in_state(\<sigma>) \<Longrightarrow> t1@t \<in> ic_traces_ending_in_state(\<sigma>') \<Longrightarrow> t2@t \<in> ic_traces_ending_in_state(\<sigma>')"
proof -
  have helper: "\<And>t1 t2 \<sigma> \<sigma>' a. t1 \<in> ic_traces_ending_in_state(\<sigma>) \<Longrightarrow> t2 \<in> ic_traces_ending_in_state(\<sigma>) \<Longrightarrow> t1@[a] \<in> ic_traces_ending_in_state(\<sigma>') \<Longrightarrow> t2@[a] \<in> ic_traces_ending_in_state(\<sigma>')"
    proof -
      fix t1 t2 \<sigma> \<sigma>' a
      assume a: "t1 \<in> ic_traces_ending_in_state(\<sigma>)"
      assume b: "t2 \<in> ic_traces_ending_in_state(\<sigma>)"
      assume "t1@[a] \<in> ic_traces_ending_in_state(\<sigma>')"
      then have "\<exists>\<sigma>''. a = (\<sigma>'',\<sigma>') \<and> (t1 = [] \<or> t1 \<in> ic_traces_ending_in_state(\<sigma>''))" by (simp add: ic_traces_ending_in_state_snoc)
      then obtain \<sigma>'' where c: "a = (\<sigma>'',\<sigma>')" and d: "t1 = [] \<or> t1 \<in> ic_traces_ending_in_state(\<sigma>'')" by blast
      from d a have "t1 \<in> ic_traces_ending_in_state(\<sigma>'')" by (auto simp add: ic_traces_ending_in_state_def)
      then have e: "\<sigma> = \<sigma>''" by (metis a trace_can_end_only_in_one_state)
      from b c e have "\<exists>\<sigma>''. a = (\<sigma>'',\<sigma>') \<and> (t2 = [] \<or> t2 \<in> ic_traces_ending_in_state(\<sigma>''))" by auto
      then show " t2@[a] \<in> ic_traces_ending_in_state(\<sigma>')" by (simp add: ic_traces_ending_in_state_snoc)
    qed
  show "\<And>t1 t2 \<sigma> \<sigma>'. t1 \<in> ic_traces_ending_in_state(\<sigma>) \<Longrightarrow> t2 \<in> ic_traces_ending_in_state(\<sigma>) \<Longrightarrow> t1@t \<in> ic_traces_ending_in_state(\<sigma>') \<Longrightarrow> t2@t \<in> ic_traces_ending_in_state(\<sigma>')"
  proof (induct t, simp, metis trace_can_end_only_in_one_state)
    fix a fix t::"'a trace" fix t1 t2 \<sigma> \<sigma>'
    assume IH: "\<And>t1 t2 \<sigma> \<sigma>'. t1 \<in> ic_traces_ending_in_state \<sigma> \<Longrightarrow> t2 \<in> ic_traces_ending_in_state \<sigma> \<Longrightarrow> t1 @ t \<in> ic_traces_ending_in_state \<sigma>' \<Longrightarrow> t2 @ t \<in> ic_traces_ending_in_state \<sigma>'"
    assume a: "(t1::'a trace) \<in> ic_traces_ending_in_state \<sigma>"
    assume b: "t2 \<in> ic_traces_ending_in_state \<sigma>"
    assume "t1 @ (a # t) \<in> ic_traces_ending_in_state \<sigma>'"
    then have c: "(t1 @ [a]) @ t \<in> ic_traces_ending_in_state \<sigma>'" by (metis append_Cons append_Nil append_assoc)
    from c have d: "ic((t1@[a])@t)" by (auto simp add: ic_traces_ending_in_state_def) 
    from d have e: "ic(t1 @ [a])" by (metis consistent_append)
    have f: "t1 @ [a] \<in> ic_traces_ending_in_state (snd a)" by (unfold ic_traces_ending_in_state_def, auto simp add: e, metis pair_collapse)
    have g: "t2 @ [a] \<in> ic_traces_ending_in_state (snd a)" by (metis a b helper f)
    from f g c have "(t2 @ [a]) @ t \<in> ic_traces_ending_in_state \<sigma>'" by (rule IH)
    then show "t2 @ (a # t) \<in> ic_traces_ending_in_state \<sigma>'" by (metis append_Cons append_Nil append_assoc)
  qed
qed
definition Inconsistent :: "'a description" where
  "Inconsistent \<equiv> {t. \<not>ic(t)}"
lemma mem_Inconsistent_iff: "t \<in> Inconsistent \<longleftrightarrow> \<not>ic(t)"
  by (auto simp add: Inconsistent_def)
lemma inconsistent_trace_extend: "\<not>ic(t) \<Longrightarrow> \<not>ic(t@t')"
  by (metis consistent_append)
lemma inconsistency_persists: "Inconsistent;P \<subseteq> Inconsistent"
proof -
  have a: "\<forall>t. t \<in> Inconsistent;P \<longrightarrow> \<not>ic(t)"
  proof (rule allI)
    fix t
    have "t \<in> Inconsistent;P \<longleftrightarrow> (\<exists>i p. i \<in> Inconsistent \<and> p \<in> P \<and> t = i@p)" by (metis elem_of_lang_concat mem_def)
    also have "\<dots> \<longleftrightarrow> (\<exists>i p. \<not>ic(i) \<and> p \<in> P \<and> t = i@p)" by (auto simp add: Inconsistent_def)
    also have "\<dots> \<longrightarrow> \<not>ic(t)" by (auto simp add: inconsistent_trace_extend)
    finally show "t \<in> Inconsistent;P \<longrightarrow> \<not>ic(t)" .
  qed
  from a have "\<forall>t. t \<in> Inconsistent;P \<longrightarrow> \<not>ic(t)" by auto
  then have "\<forall>t. t \<in> Inconsistent;P \<longrightarrow> t \<in> Inconsistent" by (simp add: mem_Inconsistent_iff)
  then show "Inconsistent;P \<subseteq> Inconsistent" by auto
qed


type_synonym 'a atom_sequence = "'a atom list"

type_synonym 'a command = "'a atom language"

fun trace_set_of_atom_sequence :: "'a atom_sequence \<Rightarrow> 'a description" where
   "trace_set_of_atom_sequence [] = skip"
 | "trace_set_of_atom_sequence (a#as) = Rep_atom(a);trace_set_of_atom_sequence(as)"
definition trace_set_of_command :: "'a command \<Rightarrow> 'a description" where
  "trace_set_of_command C \<equiv> \<Union>{P. \<exists>as \<in> C. P = trace_set_of_atom_sequence(as)}"
(* Homomorphism helping lemmas that may be of general use. *)
lemma empty_in_trace_set_of_atom_sequence: "[] \<in> trace_set_of_atom_sequence(as) \<longleftrightarrow> as = []"
  by (cases as, simp add: elem_of_skip, metis empty_elem_of_lang_concat list.simps(2) trace_of_atom_characterization trace_set_of_atom_sequence.simps(2))
lemma nonempty_in_trace_set_of_atom_sequence: "sp#t \<in> trace_set_of_atom_sequence(as) \<longleftrightarrow> (\<exists>a as'. as = a#as' \<and> [sp] \<in> Rep_atom(a) \<and> t \<in> trace_set_of_atom_sequence(as'))"
proof -
  have helper: "\<And>a sp t Q. sp#t \<in> Rep_atom(a);Q \<longleftrightarrow> [sp] \<in> Rep_atom(a) \<and> t \<in> Q"
  proof - fix a sp t Q
    have "sp#t \<in> Rep_atom(a);Q \<longleftrightarrow> ([] \<in> Rep_atom(a) \<and> sp#t \<in> Q) \<or> (\<exists>t1 t2. t = t1@t2 \<and> sp#t1 \<in> Rep_atom(a) \<and> t2 \<in> Q)" by (simp add: nonempty_elem_of_lang_concat)
    also have "\<dots> \<longleftrightarrow> (\<exists>t1 t2. t = t1@t2 \<and> sp#t1 \<in> Rep_atom(a) \<and> t2 \<in> Q)" by (metis list.simps(2) trace_of_atom_characterization)
    also have "\<dots> \<longleftrightarrow> [sp] \<in> Rep_atom(a) \<and> t \<in> Q" by (metis eq_Nil_appendI list.inject trace_of_atom_characterization)
    finally show "sp#t \<in> Rep_atom(a);Q \<longleftrightarrow> [sp] \<in> Rep_atom(a) \<and> t \<in> Q" by simp
  qed
  have a1: "sp#t \<in> trace_set_of_atom_sequence(as) \<longleftrightarrow> (\<exists>a as'. as = a#as' \<and>  sp#t \<in> Rep_atom(a);trace_set_of_atom_sequence(as'))" by (metis trace_set_of_atom_sequence.simps(1) trace_set_of_atom_sequence.simps(2) elem_of_skip neq_Nil_conv)
  have a2: "\<dots> \<longleftrightarrow> (\<exists>a as'. as = a#as' \<and> [sp] \<in> Rep_atom(a) \<and> t \<in> trace_set_of_atom_sequence(as'))" by (metis helper)
  from a1 a2 show ?thesis by simp
qed
lemma trace_set_of_atom_sequence_concat: "trace_set_of_atom_sequence(as@as') = trace_set_of_atom_sequence(as);trace_set_of_atom_sequence(as')"
  by (induct as, simp add: lang_concat_unit1, simp add: lang_concat_associative)
lemma trace_set_of_atom_sequence_interleave: "trace_set_of_command(as1\<otimes>as2) = trace_set_of_atom_sequence(as1)\<parallel>trace_set_of_atom_sequence(as2)"
proof -
  have helper: "\<And>e as as'. (\<exists>as''. as'' \<in> as\<otimes>as' \<and> e \<in> trace_set_of_atom_sequence(as'')) \<longleftrightarrow> (\<exists>t t'. t \<in> trace_set_of_atom_sequence(as) \<and> t' \<in> trace_set_of_atom_sequence(as') \<and> e \<in> t\<otimes>t')"
  proof -
    fix e
    show "\<And>as as'. (\<exists>as''. as'' \<in> as\<otimes>as' \<and> e \<in> trace_set_of_atom_sequence(as'')) \<longleftrightarrow> (\<exists>t t'. t \<in> trace_set_of_atom_sequence(as) \<and> t' \<in> trace_set_of_atom_sequence(as') \<and> e \<in> t\<otimes>t')"
    proof (induct e) fix as as'
      have "(\<exists>as''. as'' \<in> as \<otimes> as' \<and> [] \<in> trace_set_of_atom_sequence as'') \<longleftrightarrow> (\<exists>as''. as'' \<in> as \<otimes> as' \<and> as'' = [])" by (metis empty_in_trace_set_of_atom_sequence)
      also have "\<dots> \<longleftrightarrow> (as = [] \<and> as' = [])" by (metis interleave_contains_empty)
      also have "\<dots> \<longleftrightarrow> ([] \<in> trace_set_of_atom_sequence as \<and> [] \<in> trace_set_of_atom_sequence as' \<and> [] \<in> [] \<otimes> [])" by (simp, metis empty_in_trace_set_of_atom_sequence)
      also have "([] \<in> trace_set_of_atom_sequence as \<and> [] \<in> trace_set_of_atom_sequence as' \<and> [] \<in> [] \<otimes> []) \<longleftrightarrow> (\<exists>t t'. t \<in> trace_set_of_atom_sequence as \<and> t' \<in> trace_set_of_atom_sequence as' \<and> [] \<in> t \<otimes> t')" by (metis interleave_contains_empty)
      finally show "(\<exists>as''. as'' \<in> as \<otimes> as' \<and> [] \<in> trace_set_of_atom_sequence as'') \<longleftrightarrow> (\<exists>t t'. t \<in> trace_set_of_atom_sequence as \<and> t' \<in> trace_set_of_atom_sequence as' \<and> [] \<in> t \<otimes> t')" by metis
    next
      fix sp t'' as as'  (* sp stands for 'state pair' *)
      assume IH: "\<And>as as'. (\<exists>as''. as'' \<in> as\<otimes>as' \<and> t'' \<in> trace_set_of_atom_sequence as'') \<longleftrightarrow> (\<exists>tas tas'. tas \<in> trace_set_of_atom_sequence as \<and> tas' \<in> trace_set_of_atom_sequence as' \<and> t'' \<in> tas\<otimes>tas')"
      have " (\<exists>as''. as'' \<in> as\<otimes>as' \<and> sp#t'' \<in> trace_set_of_atom_sequence as'') \<longleftrightarrow> (\<exists>a as'''. a#as''' \<in> as\<otimes>as' \<and> [sp] \<in> Rep_atom(a) \<and> t'' \<in> trace_set_of_atom_sequence(as'''))" by (metis nonempty_in_trace_set_of_atom_sequence)
      also have "\<dots> \<longleftrightarrow> (\<exists>a as'''. ((\<exists>as1'. as = a#as1' \<and> as''' \<in> as1' \<otimes> as') \<or> (\<exists>as1'. as' = a#as1' \<and> as''' \<in> as \<otimes> as1')) \<and> [sp] \<in> Rep_atom(a) \<and> t'' \<in> trace_set_of_atom_sequence(as'''))" by (simp add: interleave_contains_nonempty)
      also have "\<dots> \<longleftrightarrow> (\<exists>a as1'. as = a#as1' \<and> [sp] \<in> Rep_atom(a) \<and> (\<exists>as'''.  as''' \<in> as1'\<otimes>as' \<and> t'' \<in> trace_set_of_atom_sequence(as'''))) \<or>
                        (\<exists>a as1'. as' = a#as1' \<and> [sp] \<in> Rep_atom(a) \<and> (\<exists>as'''. as''' \<in> as\<otimes>as1' \<and> t'' \<in> trace_set_of_atom_sequence(as''')))" by auto
      also have "\<dots> \<longleftrightarrow> (\<exists>a as1'. as = a#as1' \<and> [sp] \<in> Rep_atom(a) \<and> (\<exists>tp tp'. tp \<in> trace_set_of_atom_sequence as1' \<and> tp' \<in> trace_set_of_atom_sequence as' \<and> t'' \<in> tp\<otimes>tp')) \<or>
                        (\<exists>a as1'. as' = a#as1' \<and> [sp] \<in> Rep_atom(a) \<and> (\<exists>tp tp'. tp \<in> trace_set_of_atom_sequence as \<and> tp' \<in> trace_set_of_atom_sequence as1' \<and> t'' \<in> tp\<otimes>tp'))" by (auto simp add: IH)
      also have "\<dots> \<longleftrightarrow> (\<exists>tp tp'. (\<exists>a as1'. as = a#as1' \<and> [sp] \<in> Rep_atom(a) \<and> tp \<in> trace_set_of_atom_sequence as1') \<and> tp' \<in> trace_set_of_atom_sequence as' \<and> t'' \<in> tp\<otimes>tp') \<or>
                        (\<exists>tp tp'. (\<exists>a as1'. as' = a#as1' \<and> [sp] \<in> Rep_atom(a) \<and> tp' \<in> trace_set_of_atom_sequence as1') \<and> tp \<in> trace_set_of_atom_sequence as \<and> t'' \<in> tp\<otimes>tp')" by auto
      also have "\<dots> \<longleftrightarrow> (\<exists>tp tp'. sp#tp \<in> trace_set_of_atom_sequence(as) \<and> tp' \<in> trace_set_of_atom_sequence as' \<and> t'' \<in> tp\<otimes>tp') \<or>
                        (\<exists>tp tp'. sp#tp' \<in> trace_set_of_atom_sequence(as') \<and> tp \<in> trace_set_of_atom_sequence as \<and> t'' \<in> tp\<otimes>tp')" by (simp add: nonempty_in_trace_set_of_atom_sequence)
      also have "\<dots> \<longleftrightarrow> (\<exists>t t'. t \<in> trace_set_of_atom_sequence(as) \<and> t' \<in> trace_set_of_atom_sequence(as') \<and> sp#t'' \<in> t\<otimes>t')" by (auto simp add: interleave_contains_nonempty)
      finally show "(\<exists>as''. as'' \<in> as\<otimes>as' \<and> sp#t'' \<in> trace_set_of_atom_sequence as'') \<longleftrightarrow> (\<exists>tas tas'. tas \<in> trace_set_of_atom_sequence as \<and> tas' \<in> trace_set_of_atom_sequence as' \<and> sp#t'' \<in> tas \<otimes> tas')" by simp
    qed
  qed
  show "trace_set_of_command(as1\<otimes>as2) = trace_set_of_atom_sequence(as1)\<parallel>trace_set_of_atom_sequence(as2)"
  proof (unfold trace_set_of_command_def language_interleave_def)
    have b: "\<And>e. (\<exists>as''. as'' \<in> as1\<otimes>as2 \<and> e \<in> trace_set_of_atom_sequence(as'')) \<longleftrightarrow> (\<exists>t t'. t \<in> trace_set_of_atom_sequence(as1) \<and> t' \<in> trace_set_of_atom_sequence(as2) \<and> e \<in> t\<otimes>t')" by (rule helper)
    show "\<Union>{P. \<exists>as \<in> as1\<otimes>as2. P = trace_set_of_atom_sequence as} = \<Union>{r. \<exists>p q. p \<in> trace_set_of_atom_sequence as1 \<and> q \<in> trace_set_of_atom_sequence as2 \<and> r = p\<otimes>q}"
    proof -
      have "\<forall>e. e \<in> \<Union>{P. \<exists>as \<in> as1\<otimes>as2. P = trace_set_of_atom_sequence as} \<longleftrightarrow> e \<in> \<Union>{r. \<exists>p q. p \<in> trace_set_of_atom_sequence as1 \<and> q \<in> trace_set_of_atom_sequence as2 \<and> r = p\<otimes>q}"
      proof (rule allI) fix e
        have "(e \<in> \<Union>{P. \<exists>as \<in> as1\<otimes>as2. P = trace_set_of_atom_sequence as}) \<longleftrightarrow> (\<exists>as. as \<in> as1\<otimes>as2 \<and> e \<in> trace_set_of_atom_sequence as)" by auto
        also have "\<dots> \<longleftrightarrow> (\<exists>t t'. t \<in> trace_set_of_atom_sequence as1 \<and> t' \<in> trace_set_of_atom_sequence as2 \<and> e \<in> t\<otimes>t')" by (simp add: b)
        also have "\<dots> \<longleftrightarrow> (e \<in> \<Union>{r. \<exists>p q. p \<in> trace_set_of_atom_sequence as1 \<and> q \<in> trace_set_of_atom_sequence as2 \<and> r = p\<otimes>q})" by auto
        finally show "(e \<in> \<Union>{P. \<exists>as \<in> as1\<otimes>as2. P = trace_set_of_atom_sequence as}) \<longleftrightarrow> (e \<in> \<Union>{r. \<exists>p q. p \<in> trace_set_of_atom_sequence as1 \<and> q \<in> trace_set_of_atom_sequence as2 \<and> r = p\<otimes>q})" by simp
      qed
      then show ?thesis by (simp add: set_eq_iff)
    qed
  qed
qed
(* Homomorphism *)
lemma atom_homomorphism: "trace_set_of_command({[a]}) = Rep_atom(a)"
  by (unfold trace_set_of_command_def, auto simp add: lang_concat_unit1 lang_concat_unit2)
lemma skip_homomorphism: "trace_set_of_command(skip) = skip"
  by (unfold trace_set_of_command_def skip_def, auto simp add: elem_of_skip)
lemma seq_homomorphism: "trace_set_of_command(C;C') = trace_set_of_command(C);trace_set_of_command(C')"
proof -
  have "trace_set_of_command(C;C') = \<Union>{P. \<exists>as. as \<in> (C;C') \<and> P = trace_set_of_atom_sequence as}" by (auto simp add: trace_set_of_command_def)
  also have "\<dots> = \<Union>{P. \<exists>as. (\<exists>as1 as2. as1 \<in> C \<and> as2 \<in> C' \<and> as = as1@as2) \<and> P = trace_set_of_atom_sequence as}" by (simp add: elem_of_lang_concat)
  also have "\<dots> = \<Union>{P. \<exists>as1 as2. as1 \<in> C \<and> as2 \<in> C' \<and> P = trace_set_of_atom_sequence(as1@as2)}" by auto
  also have "\<dots> = \<Union>{P. \<exists>as1 as2. as1 \<in> C \<and> as2 \<in> C' \<and> P = trace_set_of_atom_sequence(as1);trace_set_of_atom_sequence(as2)}" by (simp add: trace_set_of_atom_sequence_concat)
  also have "\<dots> = \<Union>{P. \<exists>as1 as2. as1 \<in> C \<and> as2 \<in> C' \<and> P = {r. \<exists>t1 t2. t1 \<in> trace_set_of_atom_sequence(as1) \<and> t2 \<in> trace_set_of_atom_sequence(as2) \<and> r = t1@t2}}" by (simp add: language_concat_def)
  also have "\<dots> = \<Union>{P. P = {r. \<exists>t1 t2. t1 \<in> trace_set_of_command(C) \<and> t2 \<in> trace_set_of_command(C') \<and> r = t1@t2}}" by (auto simp add: trace_set_of_command_def)
  also have "\<dots> = \<Union>{P. P = {r. r \<in> trace_set_of_command(C);trace_set_of_command(C')}}" by (simp add: elem_of_lang_concat)
  also have "\<dots> = trace_set_of_command(C);trace_set_of_command(C')" by auto
  finally show "trace_set_of_command(C;C') = trace_set_of_command(C);trace_set_of_command(C')" by simp
qed
lemma choice_homomorphism: "trace_set_of_command(\<Union>Y) = \<Union>{P. \<exists>C \<in> Y. P = trace_set_of_command(C)}"
  by (auto simp add: trace_set_of_command_def)
lemma binary_choice_homomorphism: "trace_set_of_command(C \<union> C') = trace_set_of_command(C) \<union> trace_set_of_command(C')"
  by (auto simp add: trace_set_of_command_def)
lemma iter_homomorphism: "trace_set_of_command(iter C n) = iter (trace_set_of_command(C)) n"
  by (induct n, simp, metis skip_homomorphism, simp, metis seq_homomorphism)
lemma kleene_homomorphism: "trace_set_of_command(C\<^sup>\<star>) = trace_set_of_command(C)\<^sup>\<star>"
proof (unfold kleene_star_def)
  have "trace_set_of_command(\<Union>{Q. \<exists>n. Q = iter C n}) = \<Union>{Q. \<exists>n. Q = trace_set_of_command(iter C n)}" by (auto simp add: choice_homomorphism)
  also have "\<dots> = \<Union>{Q. \<exists>n. Q = (iter (trace_set_of_command C) n)}" by (simp add: iter_homomorphism)
  finally show "trace_set_of_command(\<Union>{Q. \<exists>n. Q = iter C n}) = \<Union>{Q. \<exists>n. Q = (iter (trace_set_of_command C) n)}" by simp
qed
lemma conc_homomorphism: "trace_set_of_command(C\<parallel>C') = trace_set_of_command(C)\<parallel>trace_set_of_command(C')"
proof -
  have "trace_set_of_command(C\<parallel>C') = \<Union>{P. \<exists>as. as \<in> (C\<parallel>C') \<and> P = trace_set_of_atom_sequence as}" by (auto simp add: trace_set_of_command_def)
  also have "\<dots> = \<Union>{P. \<exists>as. (\<exists>as1 as2. as1 \<in> C \<and> as2 \<in> C' \<and> as \<in> as1\<otimes>as2) \<and> P = trace_set_of_atom_sequence as}" by (simp add: elem_of_lang_interleave)
  also have "\<dots> = {t. \<exists>as1 as2. as1 \<in> C \<and> as2 \<in> C' \<and> t \<in> \<Union>{P. \<exists>as \<in> as1\<otimes>as2. P = trace_set_of_atom_sequence(as)}}" by auto
  also have "\<dots> = {t. \<exists>as1 as2. as1 \<in> C \<and> as2 \<in> C' \<and> t \<in> trace_set_of_command(as1\<otimes>as2)}" by (simp add: trace_set_of_command_def)
  also have "\<dots> = {t. \<exists>as1 as2. as1 \<in> C \<and> as2 \<in> C' \<and> t \<in> trace_set_of_atom_sequence(as1)\<parallel>trace_set_of_atom_sequence(as2)}" by (auto simp add: trace_set_of_atom_sequence_interleave)
  also have "\<dots> = {t. \<exists>as1 as2. as1 \<in> C \<and> as2 \<in> C' \<and> (\<exists>t1 t2. t1 \<in> trace_set_of_atom_sequence(as1) \<and> t2 \<in> trace_set_of_atom_sequence(as2) \<and> t \<in> t1\<otimes>t2)}" by (simp add: elem_of_lang_interleave)
  also have "\<dots> = {t. \<exists>t1 t2. t1 \<in> trace_set_of_command(C) \<and> t2 \<in> trace_set_of_command(C') \<and> t \<in> t1\<otimes>t2}" by (auto simp add: trace_set_of_command_def)
  also have "\<dots> = trace_set_of_command(C)\<parallel>trace_set_of_command(C')" by (auto simp add: language_interleave_def)
  finally show "trace_set_of_command(C\<parallel>C') = trace_set_of_command(C)\<parallel>trace_set_of_command(C')" by simp
qed

end
