(*  Title:       Languages.thy
    Author:      Stephan van Staden, ETH Zurich, Switzerland
    Description: Formal languages with interleaving.
    Copyright    2012 ETH Zurich
*)

theory Languages
  imports Interleaving
begin

declare [[ smt_solver = remote_z3]]
declare [[ smt_timeout = 60 ]]
declare [[ z3_options = "-memory:500" ]]

type_synonym 'a language = "'a list set"

definition bottom :: "'a language" ("\<bottom>") where
  "\<bottom> \<equiv> {}"

definition top :: "'a language" ("\<top>") where
  "\<top> \<equiv> {p. True}"

definition skip :: "'a language" where
  "skip \<equiv> {[]}"

definition language_concat :: "'a language \<Rightarrow> 'a language \<Rightarrow> 'a language" (infix ";" 67) where
  "P;Q \<equiv> {r. \<exists>p q. p \<in> P \<and> q \<in> Q \<and> r = p@q}"

fun iter :: "'a language \<Rightarrow> nat \<Rightarrow> 'a language" where
   "iter P 0 = skip"
 | "iter P (Suc n) = P ; (iter P n)"

definition kleene_star :: "'a language \<Rightarrow> 'a language" ("_\<^sup>\<star>" [68] 69) where
  "P\<^sup>\<star> \<equiv> \<Union>{Q. \<exists>n::nat. Q = iter P n}"

definition language_interleave :: "'a language \<Rightarrow> 'a language \<Rightarrow> 'a language" (infix "\<parallel>" 67) where
  "P\<parallel>Q \<equiv> \<Union>{r. \<exists>p q. p \<in> P \<and> q \<in> Q \<and> r = p\<otimes>q}"

(* Simple lemmas about membership. *)

lemma elem_of_skip: "s \<in> skip \<longleftrightarrow> s = []"
  by (metis singleton_iff skip_def)

lemma elem_of_lang_concat: "s \<in> P;Q \<longleftrightarrow> (\<exists>p q. p \<in> P \<and> q \<in> Q \<and> s = p@q)"
  by (simp add: Collect_def language_concat_def mem_def)

lemma empty_elem_of_lang_concat: "[] \<in> P;Q \<longleftrightarrow> [] \<in> P \<and> [] \<in> Q"
  by (metis append_is_Nil_conv elem_of_lang_concat)

lemma append_is_nonempty: "(e#s = p@q) \<longleftrightarrow> (p = [] \<and> e#s = q) \<or> (\<exists>t. s = t@q \<and> p = e#t)"
  by (metis Cons_eq_append_conv)

lemma nonempty_elem_of_lang_concat: "e#s \<in> P;Q \<longleftrightarrow> ([] \<in> P \<and> e#s \<in> Q) \<or> (\<exists>t t'. s = t@t' \<and> e#t \<in> P \<and> t' \<in> Q)"
proof -
  have "e#s \<in> P;Q \<longleftrightarrow> (\<exists>p q. p \<in> P \<and> q \<in> Q \<and> e#s = p@q)" by (simp add: elem_of_lang_concat)
  also have "\<dots> \<longleftrightarrow> (\<exists>p q. p \<in> P \<and> q \<in> Q \<and> ((p = [] \<and> e#s = q) \<or> (\<exists>t. s = t@q \<and> p = e#t)))" by (simp add: append_is_nonempty)
  also have "\<dots> \<longleftrightarrow> ([] \<in> P \<and> e#s \<in> Q) \<or> (\<exists>t t'. s = t@t' \<and> e#t \<in> P \<and> t' \<in> Q)" by metis
  finally show ?thesis by simp
qed

lemma elem_of_kleene_star: "s \<in> P\<^sup>\<star> \<longleftrightarrow> (\<exists>n::nat. s \<in> iter P n)"
proof -
  have "s \<in> P\<^sup>\<star> \<longleftrightarrow> (\<exists>Q \<in> {Q. \<exists>n::nat. Q = iter P n}. s \<in> Q)" by (simp add: kleene_star_def)
  also have "\<dots> \<longleftrightarrow> (\<exists>Q n::nat. Q = iter P n \<and> s \<in> Q)" by simp
  also have "\<dots> \<longleftrightarrow> (\<exists>n::nat. s \<in> iter P n)" by metis
  finally show ?thesis by metis
qed

lemma elem_of_lang_interleave: "s \<in> P\<parallel>Q \<longleftrightarrow> (\<exists>p q. p \<in> P \<and> q \<in> Q \<and> s \<in> p\<otimes>q)"
proof -
  have "s \<in> P\<parallel>Q \<longleftrightarrow> s \<in> \<Union>{r. \<exists>p q. p \<in> P \<and> q \<in> Q \<and> r = p\<otimes>q}" by (simp add: language_interleave_def)
  also have "\<dots> \<longleftrightarrow> (\<exists>r \<in> {r. \<exists>p q. p \<in> P \<and> q \<in> Q \<and> r = p\<otimes>q}. s \<in> r)" by simp
  also have "\<dots> \<longleftrightarrow> (\<exists>r. (\<exists>p q. p \<in> P \<and> q \<in> Q \<and> r = p\<otimes>q) \<and> s \<in> r)" by simp
  also have "\<dots> \<longleftrightarrow> (\<exists>p q. p \<in> P \<and> q \<in> Q \<and> s \<in> p\<otimes>q)" by metis
  finally show ?thesis .
qed

lemma empty_elem_of_lang_interleave: "[] \<in> P\<parallel>Q \<longleftrightarrow> [] \<in> P \<and> [] \<in> Q"
  by (metis elem_of_lang_interleave interleave_contains_empty)

lemma nonempty_elem_of_lang_interleave: "e#s \<in> (P\<parallel>Q) \<longleftrightarrow> (\<exists>p' q. e#p' \<in> P \<and> q \<in> Q \<and> s \<in> p'\<otimes>q) \<or> (\<exists>p q'. p \<in> P \<and> e#q' \<in> Q \<and> s \<in> p\<otimes>q')"
proof -
  have "e#s \<in> P\<parallel>Q \<longleftrightarrow> (\<exists>p q. p \<in> P \<and> q \<in> Q \<and> e#s \<in> p\<otimes>q)" by (metis elem_of_lang_interleave)
  also have "\<dots> \<longleftrightarrow> (\<exists>p q. p \<in> P \<and> q \<in> Q \<and> ((\<exists>p'. p = e#p' \<and> s \<in> p'\<otimes>q) \<or> (\<exists>q'. q = e#q' \<and> s \<in> p\<otimes>q')))" by (simp add: interleave_contains_nonempty)
  also have "\<dots> \<longleftrightarrow> (\<exists>p' q. e#p' \<in> P \<and> q \<in> Q \<and> s \<in> p'\<otimes>q) \<or> (\<exists>p q'. p \<in> P \<and> e#q' \<in> Q \<and> s \<in> p\<otimes>q')" by auto
  finally show ?thesis by metis
qed

(* The table of properties. *)

lemma choice_unit: "P \<union> \<bottom> = P"
  by (metis Un_empty_right bottom_def)

lemma choice_zero: "P \<union> \<top> = \<top>"
  by (metis UNIV_def Un_UNIV_right top_def)

lemma intersect_unit: "P \<inter> \<top> = P"
  by (metis Int_UNIV_right UNIV_def top_def)

lemma intersect_zero: "P \<inter> \<bottom> = \<bottom>"
  by (metis Int_empty_right bottom_def)

lemma lang_concat_associative: "P;(Q;R) = (P;Q);R"
proof -
  have alternative: "\<forall>t. t \<in> P;(Q;R) \<longleftrightarrow> t \<in> (P;Q);R"
  proof (rule allI)
    fix t
    have "t \<in> P;(Q;R) \<longleftrightarrow> (\<exists>p s1. p \<in> P \<and> (\<exists>q r. q \<in> Q \<and> r \<in> R \<and>  s1 = q@r) \<and> t = p@s1)" by (simp add: elem_of_lang_concat)
    also have "\<dots> \<longleftrightarrow> (\<exists>p q r. p \<in> P \<and> q \<in> Q \<and> r \<in> R \<and> (\<exists>s1. s1 = q@r \<and> t = p@s1))" by metis
    also have "\<dots> \<longleftrightarrow> (\<exists>p q r. p \<in> P \<and> q \<in> Q \<and> r \<in> R \<and> t = (p@q)@r)" by simp
    also have "\<dots> \<longleftrightarrow> (\<exists>p q r. p \<in> P \<and> q \<in> Q \<and> r \<in> R \<and> (\<exists>s2. s2 = p@q \<and> t = s2@r))" by metis
    also have "\<dots> \<longleftrightarrow> (\<exists>s2 r. (\<exists>p q. p \<in> P \<and> q \<in> Q \<and> s2 = p@q) \<and> r \<in> R \<and> t = s2@r)" by metis
    also have "\<dots> \<longleftrightarrow> t \<in> (P;Q);R" by (simp add: elem_of_lang_concat)
    finally show "t \<in> P;(Q;R) \<longleftrightarrow> t \<in> (P;Q);R" .
  qed
  then show ?thesis by (simp add: set_eq_iff)
qed

lemma lang_concat_unit1: "skip;P = P"
  by (simp add: language_concat_def skip_def)

lemma lang_concat_unit2: "P;skip = P"
  by (simp add: language_concat_def skip_def)

lemma lang_concat_zero1: "\<bottom>;P = \<bottom>"
  by (simp add: language_concat_def bottom_def)

lemma lang_concat_zero2: "P;\<bottom> = \<bottom>"
  by (simp add: language_concat_def bottom_def)

lemma lang_interleave_commutative: "P\<parallel>Q = Q\<parallel>P"
proof -
  have a: "(\<lambda>r. (\<exists>p q. p \<in> P \<and> q \<in> Q \<and> r = q\<otimes>p)) = (\<lambda>r. (\<exists>q p. q \<in> Q \<and> p \<in> P \<and> r = q\<otimes>p))" by metis
  have b: "P\<parallel>Q = \<Union>{r. \<exists>p q. p \<in> P \<and> q \<in> Q \<and> r = p\<otimes>q}" by (simp add: language_interleave_def)
  also have c: "\<dots> = \<Union>{r. \<exists>p q. p \<in> P \<and> q \<in> Q \<and> r = q\<otimes>p}" by (simp add: interleave_commutative)
  from a have d: "\<Union>{r. \<exists>p q. p \<in> P \<and> q \<in> Q \<and> r = q\<otimes>p} = \<Union>{r. \<exists>q p. q \<in> Q \<and> p \<in> P \<and> r = q\<otimes>p}" by metis
  have e: "\<Union>{r. \<exists>q p. q \<in> Q \<and> p \<in> P \<and> r = q\<otimes>p} = Q\<parallel>P" by (simp add: language_interleave_def)
  from b c d e show ?thesis by metis
qed

lemma lang_interleave_associative: "P\<parallel>(Q\<parallel>R) = (P\<parallel>Q)\<parallel>R"
proof -
  have alternative: "\<forall>t. t \<in> P\<parallel>(Q\<parallel>R) \<longleftrightarrow> t \<in> (P\<parallel>Q)\<parallel>R"
  proof (rule allI)
    fix t
    have f1: "t \<in> P\<parallel>(Q\<parallel>R) \<longleftrightarrow> (\<exists>p u. p \<in> P \<and> (\<exists>q r. q \<in> Q \<and> r \<in> R \<and> u \<in> q\<otimes>r) \<and> t \<in> p\<otimes>u)" by (simp add: elem_of_lang_interleave)
    have f2: "\<dots> \<longleftrightarrow> (\<exists>p q r. p \<in> P \<and> q \<in> Q \<and> r \<in> R \<and> (\<exists>u. u \<in> q\<otimes>r \<and> t \<in> p\<otimes>u))" by metis
    have f3: "\<dots> \<longleftrightarrow> (\<exists>p q r. p \<in> P \<and> q \<in> Q \<and> r \<in> R \<and> (\<exists>s. s \<in> p\<otimes>q \<and> t \<in> s\<otimes>r))"
    proof - (* This is really really really annoying. *)
      have a: "(\<exists>p q r. p \<in> P \<and> q \<in> Q \<and> r \<in> R \<and> (\<exists>u. u \<in> q\<otimes>r \<and> t \<in> p\<otimes>u)) \<longrightarrow> (\<exists>p q r. p \<in> P \<and> q \<in> Q \<and> r \<in> R \<and> (\<exists>s. s \<in> p\<otimes>q \<and> t \<in> s\<otimes>r))"
      proof (rule impI)
        assume "\<exists>p q r. p \<in> P \<and> q \<in> Q \<and> r \<in> R \<and> (\<exists>u. u \<in> q\<otimes>r \<and> t \<in> p\<otimes>u)"
        then obtain p q r where a1: "p \<in> P \<and> q \<in> Q \<and> r \<in> R \<and> (\<exists>u. u \<in> q\<otimes>r \<and> t \<in> p\<otimes>u)" by blast
        from a1 have "p \<in> P \<and> q \<in> Q \<and> r \<in> R \<and> (\<exists>s. s \<in> p\<otimes>q \<and> t \<in> s\<otimes>r)" by (metis interleave_assoc)
        then show "\<exists>p q r. p \<in> P \<and> q \<in> Q \<and> r \<in> R \<and> (\<exists>s. s \<in> p\<otimes>q \<and> t \<in> s\<otimes>r)" by metis
      qed
      have b: "(\<exists>p q r. p \<in> P \<and> q \<in> Q \<and> r \<in> R \<and> (\<exists>s. s \<in> p\<otimes>q \<and> t \<in> s\<otimes>r)) \<longrightarrow> (\<exists>p q r. p \<in> P \<and> q \<in> Q \<and> r \<in> R \<and> (\<exists>u. u \<in> q\<otimes>r \<and> t \<in> p\<otimes>u))"
      proof (rule impI)
        assume "\<exists>p q r. p \<in> P \<and> q \<in> Q \<and> r \<in> R \<and> (\<exists>s. s \<in> p\<otimes>q \<and> t \<in> s\<otimes>r)"
        then obtain p q r where b1: "p \<in> P \<and> q \<in> Q \<and> r \<in> R \<and> (\<exists>s. s \<in> p\<otimes>q \<and> t \<in> s\<otimes>r)" by blast
        from b1 have "p \<in> P \<and> q \<in> Q \<and> r \<in> R \<and> (\<exists>u. u \<in> q\<otimes>r \<and> t \<in> p\<otimes>u)" by (metis interleave_assoc)
        then show "\<exists>p q r. p \<in> P \<and> q \<in> Q \<and> r \<in> R \<and> (\<exists>u. u \<in> q\<otimes>r \<and> t \<in> p\<otimes>u)" by metis
      qed
      have obvious: "\<And>Pr Qr. Pr \<longrightarrow> Qr \<Longrightarrow> Qr \<longrightarrow> Pr \<Longrightarrow> Pr \<longleftrightarrow> Qr" by metis
      from a b show ?thesis by (rule obvious)
    qed
    also have f4: "\<dots> \<longleftrightarrow> (\<exists>s r. (\<exists>p q. p \<in> P \<and> q \<in> Q \<and> s \<in> p\<otimes>q) \<and> r \<in> R \<and> t \<in> s\<otimes>r)" by metis
    also have f5: "\<dots> \<longleftrightarrow> t \<in> (P\<parallel>Q)\<parallel>R" by (simp add: elem_of_lang_interleave)
    from f1 f2 f3 f4 f5 show "t \<in> P\<parallel>(Q\<parallel>R) \<longleftrightarrow> t \<in> (P\<parallel>Q)\<parallel>R" by simp
  qed
  then show ?thesis by (simp add: set_eq_iff)
qed

lemma lang_interleave_unit1: "skip\<parallel>P = P"
proof -
  have alternative: "\<forall>p. p \<in> skip\<parallel>P \<longleftrightarrow> p \<in> P"
  proof (rule allI)
    fix p
    have a: "p \<in> skip\<parallel>P \<longleftrightarrow> (\<exists>s p'. s \<in> skip \<and> p' \<in> P \<and> p \<in> s\<otimes>p')" by (simp add: elem_of_lang_interleave)
    also have b: "\<dots> \<longleftrightarrow> p \<in> P" by (simp add: elem_of_skip elem_interleave_with_empty1)
    finally show "p \<in> skip\<parallel>P \<longleftrightarrow> p \<in> P" .
  qed
  then show ?thesis by (simp add: set_eq_iff)
qed

lemma lang_interleave_unit2: "P\<parallel>skip = P"
  by (simp add: lang_interleave_unit1 lang_interleave_commutative)

lemma lang_interleave_zero1: "\<bottom>\<parallel>P = \<bottom>"
  by (metis bottom_def elem_of_lang_interleave ex_in_conv)

lemma lang_interleave_zero2: "P\<parallel>\<bottom> = \<bottom>"
  by (simp add: lang_interleave_zero1 lang_interleave_commutative)

(* Distribution laws and some consequences. *)

lemma lang_concat_distr_Union1: "\<Union>X;P = \<Union>{Q. \<exists>R\<in>X. Q = R;P}"
proof -
  have alternative: "\<forall>s. s \<in> \<Union>X;P \<longleftrightarrow> s \<in> \<Union>{Q. \<exists>R\<in>X. Q = R;P}"
  proof (rule allI)
    fix s
    have "s \<in> \<Union>X;P \<longleftrightarrow> (\<exists>q p. q \<in> \<Union>X \<and> p \<in> P \<and> s = q@p)" by (simp add: elem_of_lang_concat)
    also have "\<dots> \<longleftrightarrow> (\<exists>q p. (\<exists>R \<in> X. q \<in> R) \<and> p \<in> P \<and> s = q@p)" by simp
    also have "\<dots> \<longleftrightarrow> (\<exists>R \<in> X. \<exists>q p. q \<in> R  \<and> p \<in> P \<and> s = q@p)" by metis
    also have "\<dots> \<longleftrightarrow> (\<exists>R \<in> X. s \<in> R;P)" by (simp add: elem_of_lang_concat)
    also have "\<dots> \<longleftrightarrow> (\<exists>Q. \<exists>R\<in>X. Q = R;P \<and> s \<in> Q)" by metis
    also have "\<dots> \<longleftrightarrow> s \<in> \<Union>{Q. \<exists>R\<in>X. Q = R;P}" by simp
    finally show "s \<in> \<Union>X;P \<longleftrightarrow> s \<in> \<Union>{Q. \<exists>R\<in>X. Q = R;P}" .
  qed
  then show ?thesis by (simp add: set_eq_iff)
qed

lemma union_Union_relationship: "A \<union> B = \<Union>{C. C = A \<or> C = B}"
  by auto

lemma lang_concat_distr_union1: "(P \<union> Q);R = P;R \<union> Q;R"
proof -
  have "\<Union>{S. S = P \<or> S = Q};R = P;R \<union> Q;R" by (simp add: lang_concat_distr_Union1, auto)
  then show ?thesis by (metis union_Union_relationship)
qed

lemma lang_concat_monotone1: "P \<subseteq> Q \<Longrightarrow> P;R \<subseteq> Q;R"
  by (metis lang_concat_distr_union1 subset_Un_eq)

lemma lang_concat_distr_Union2: "P;\<Union>X = \<Union>{Q. \<exists>R\<in>X. Q = P;R}"
proof -
  have alternative: "\<forall>s. s \<in> P;\<Union>X \<longleftrightarrow> s \<in> \<Union>{Q. \<exists>R\<in>X. Q = P;R}"
  proof (rule allI)
    fix s
    have "s \<in> P;\<Union>X \<longleftrightarrow> (\<exists>p q. p \<in> P \<and> q \<in> \<Union>X \<and> s = p@q)" by (simp add: elem_of_lang_concat)
    also have "\<dots> \<longleftrightarrow> (\<exists>p q. p \<in> P \<and> (\<exists>R \<in> X. q \<in> R) \<and> s = p@q)" by simp
    also have "\<dots> \<longleftrightarrow> (\<exists>R \<in> X. \<exists>p q. p \<in> P \<and> q \<in> R \<and> s = p@q)" by metis
    also have "\<dots> \<longleftrightarrow> (\<exists>R \<in> X. s \<in> P;R)" by (simp add: elem_of_lang_concat)
    also have "\<dots> \<longleftrightarrow> (\<exists>Q. \<exists>R\<in>X. Q = P;R \<and> s \<in> Q)" by metis
    also have "\<dots> \<longleftrightarrow> s \<in> \<Union>{Q. \<exists>R\<in>X. Q = P;R}" by simp
    finally show "s \<in> P;\<Union>X \<longleftrightarrow> s \<in> \<Union>{Q. \<exists>R\<in>X. Q = P;R}" .
  qed
  then show ?thesis by (simp add: set_eq_iff)
qed

lemma lang_concat_distr_union2: "P;(Q \<union> R) = P;Q \<union> P;R"
proof -
  have "P;\<Union>{S. S = Q \<or> S = R} = P;Q \<union> P;R" by (simp add: lang_concat_distr_Union2, auto)
  then show ?thesis by (metis union_Union_relationship)
qed

lemma lang_concat_monotone2: "Q \<subseteq> R \<Longrightarrow> P;Q \<subseteq> P;R"
  by (metis lang_concat_distr_union2 subset_Un_eq)

lemma lang_interleave_distr_Union1: "\<Union>X\<parallel>P = \<Union>{Q. \<exists>R\<in>X. Q = R\<parallel>P}"
proof -
  have alternative: "\<forall>s. s \<in> \<Union>X\<parallel>P \<longleftrightarrow> s \<in> \<Union>{Q. \<exists>R\<in>X. Q = R\<parallel>P}"
  proof (rule allI)
    fix s
    have "s \<in> \<Union>X\<parallel>P \<longleftrightarrow> (\<exists>q p. q \<in> \<Union>X \<and> p \<in> P \<and> s \<in> q\<otimes>p)" by (simp add: elem_of_lang_interleave)
    also have "\<dots> \<longleftrightarrow> (\<exists>q p. (\<exists>R \<in> X. q \<in> R) \<and> p \<in> P \<and> s \<in> q\<otimes>p)" by simp
    also have "\<dots> \<longleftrightarrow> (\<exists>R \<in> X. \<exists>q p. q \<in> R  \<and> p \<in> P \<and> s \<in> q\<otimes>p)" by metis
    also have "\<dots> \<longleftrightarrow> (\<exists>R \<in> X. s \<in> R\<parallel>P)" by (simp add: elem_of_lang_interleave)
    also have "\<dots> \<longleftrightarrow> (\<exists>Q. \<exists>R\<in>X. Q = R\<parallel>P \<and> s \<in> Q)" by metis
    also have "\<dots> \<longleftrightarrow> s \<in> \<Union>{Q. \<exists>R\<in>X. Q = R\<parallel>P}" by simp
    finally show "s \<in> \<Union>X\<parallel>P \<longleftrightarrow> s \<in> \<Union>{Q. \<exists>R\<in>X. Q = R\<parallel>P}" .
  qed
  then show ?thesis by (simp add: set_eq_iff)
qed

lemma lang_interleave_distr_union1: "(P \<union> Q)\<parallel>R = P\<parallel>R \<union> Q\<parallel>R"
proof -
  have "\<Union>{S. S = P \<or> S = Q}\<parallel>R = P\<parallel>R \<union> Q\<parallel>R" by (simp add: lang_interleave_distr_Union1, auto)
  then show ?thesis by (metis union_Union_relationship)
qed

lemma lang_interleave_monotone1: "P \<subseteq> Q \<Longrightarrow> P\<parallel>R \<subseteq> Q\<parallel>R"
  by (metis lang_interleave_distr_union1 subset_Un_eq)

lemma lang_interleave_distr_Union2: "P\<parallel>\<Union>X = \<Union>{Q. \<exists>R\<in>X. Q = P\<parallel>R}"
proof -
  have "P\<parallel>\<Union>X = \<Union>X\<parallel>P" by (simp add: lang_interleave_commutative)
  also have "\<dots> = \<Union>{Q. \<exists>R\<in>X. Q = R\<parallel>P}" by (simp add: lang_interleave_distr_Union1)
  also have "\<dots> = \<Union>{Q. \<exists>R\<in>X. Q = P\<parallel>R}" by (simp add: lang_interleave_commutative)
  finally show ?thesis .
qed

lemma lang_interleave_distr_union2: "P\<parallel>(Q \<union> R) = P\<parallel>Q \<union> P\<parallel>R"
  by (metis lang_interleave_distr_union1 lang_interleave_commutative)

lemma lang_interleave_monotone2: "Q \<subseteq> R \<Longrightarrow> P\<parallel>Q \<subseteq> P\<parallel>R"
  by (metis lang_interleave_distr_union2 subset_Un_eq)

lemma lang_intersect_distr_Union1: "\<Union>X \<inter> P = \<Union>{Q. \<exists>R\<in>X. Q = R \<inter> P}"
  by auto

lemma lang_intersect_distr_union1: "(P \<union> Q) \<inter> R = (P \<inter> R) \<union> (Q \<inter> R)"
  by auto

lemma lang_intersect_monotone1: "P \<subseteq> Q \<Longrightarrow> P \<inter> R \<subseteq> Q \<inter> R"
  by auto

lemma lang_union_distr_Union1: "X \<noteq> {} \<Longrightarrow> \<Union>X \<union> P = \<Union>{Q. \<exists>R\<in>X. Q = R \<union> P}"
  by auto

lemma lang_union_distr_union1: "(P \<union> Q) \<union> R = (P \<union> R) \<union> (Q \<union> R)"
  by auto

lemma lang_union_distr_Union2: "X \<noteq> {} \<Longrightarrow> P \<union> \<Union>X = \<Union>{Q. \<exists>R\<in>X. Q = P \<union> R}"
  by auto

lemma lang_union_distr_union2: "P \<union> (Q \<union> R) = (P \<union> Q) \<union> (P \<union> R)"
  by auto

lemma lang_union_monotone1: "P \<subseteq> Q \<Longrightarrow> P \<union> R \<subseteq> Q \<union> R"
  by auto

lemma lang_union_monotone2: "P \<subseteq> Q \<Longrightarrow> R \<union> P \<subseteq> R \<union> Q"
  by auto

(* The Kleene star laws. *)

lemma skip_included_in_kleene_star: "skip \<subseteq> P\<^sup>\<star>"
  by (metis Languages.iter.simps(1) elem_of_kleene_star subsetI)

lemma kleene1part: "P;P\<^sup>\<star> \<subseteq> P\<^sup>\<star>"
proof -
  have "\<forall>Q. (\<exists>n::nat. Q = iter P (n+1)) \<longrightarrow> (\<exists>n::nat. Q = iter P n)" by metis
  then have a: "{Q. \<exists>n::nat. Q = iter P (n+1)} \<subseteq> {Q. \<exists>n::nat. Q = iter P n}" by (simp add: Collect_mono)
  have "P;P\<^sup>\<star> = P;\<Union>{R. \<exists>n::nat. R = iter P n}" by (simp add: kleene_star_def)
  also have "\<dots> = \<Union>{Q. \<exists>R\<in>{R. \<exists>n::nat. R = iter P n}. Q = P;R}" by (simp add: lang_concat_distr_Union2)
  also have "\<dots> = \<Union>{Q. \<exists>R n::nat. R = iter P n \<and> Q = P;R}" by simp
  also have "\<dots> = \<Union>{Q. \<exists>n::nat. Q = P;(iter P n)}" by auto
  also have "\<dots> = \<Union>{Q. \<exists>n::nat. Q = iter P (n+1)}" by simp
  also from a have "\<dots> \<subseteq> \<Union>{Q. \<exists>n::nat. Q = iter P n}" by (rule Union_mono)
  also have "\<dots> = P\<^sup>\<star>" by (simp add: kleene_star_def)
  finally show ?thesis .
qed

lemma kleene1: "skip \<union> P;P\<^sup>\<star> \<subseteq> P\<^sup>\<star>"
  by (metis skip_included_in_kleene_star kleene1part le_sup_iff)

lemma kleene2: "P \<union> (Q;R) \<subseteq> R \<Longrightarrow> Q\<^sup>\<star>;P \<subseteq> R"
proof -
  assume a: "P \<union> (Q;R) \<subseteq> R"
  have b: "\<forall>n::nat. (iter Q n);P \<subseteq> R"
  proof (rule allI)
    fix n
    show "(iter Q n);P \<subseteq> R"
    proof (induct n)
      have "(iter Q 0);P = P" by (metis Languages.iter.simps(1) lang_concat_unit1)
      then show "(iter Q 0);P \<subseteq> R" by (metis a le_supE)
    next
      fix n
      assume c: "(iter Q n);P \<subseteq> R"
      have d: "(iter Q (Suc n));P = Q;((iter Q n);P)" by (metis Languages.iter.simps(2) lang_concat_associative)
      also have e: "\<dots> \<subseteq> R" by (metis (no_types) Un_upper2 a c lang_concat_monotone2 xt1(6))
      show "(iter Q (Suc n));P \<subseteq> R" by (metis calculation e)
    qed
  qed
  have "\<forall>T. (\<exists>n::nat. T = (iter Q n);P) \<longrightarrow> T \<subseteq> R"
  proof (rule allI)
    fix T
    show "(\<exists>n. T = iter Q n ; P) \<longrightarrow> T \<subseteq> R"
    proof (rule impI)
      assume "\<exists>n. T = iter Q n ; P"
      then obtain n where f: "T = iter Q n ; P" by blast
      from b f show "T \<subseteq> R" by metis
    qed
  qed
  then have c: "{T. \<exists>n::nat. T = (iter Q n);P} \<subseteq> {T. T \<subseteq> R}" by (simp add: Collect_mono)
  have "Q\<^sup>\<star>;P = \<Union>{S. \<exists>n::nat. S = iter Q n};P" by (simp add: kleene_star_def)
  also have "\<dots> = \<Union>{T. \<exists>S\<in>{S. \<exists>n::nat. S = iter Q n}. T = S;P}" by (simp add: lang_concat_distr_Union1)
  also have "\<dots> = \<Union>{T. \<exists>S n::nat. S = iter Q n \<and> T = S;P}" by simp
  also have "\<dots> = \<Union>{T. \<exists>n::nat. T = (iter Q n);P}" by auto
  also from c have "\<dots> \<subseteq> \<Union>{T. T \<subseteq> R}" by (rule Union_mono)
  also have "\<dots> \<subseteq> R" by (metis Pow_def Union_Pow_eq subset_refl)
  finally show "Q\<^sup>\<star>;P \<subseteq> R" .
qed

lemma iteration_helper: "P;(iter P n) = (iter P n);P"
proof (induct n)
  show "P;(iter P 0) = (iter P 0);P" by (metis Languages.iter.simps(1) lang_concat_unit1 lang_concat_unit2)
next
  fix n
  assume IH: "P;(iter P n) = (iter P n);P"
  show "P;(iter P (Suc n)) = (iter P (Suc n));P" by (metis IH Languages.iter.simps(2) lang_concat_associative)
qed

lemma kleene3part: "P\<^sup>\<star>;P \<subseteq> P\<^sup>\<star>"
proof -
  have "\<forall>Q. (\<exists>n::nat. Q = iter P (n+1)) \<longrightarrow> (\<exists>n::nat. Q = iter P n)" by metis
  then have a: "{Q. \<exists>n::nat. Q = iter P (n+1)} \<subseteq> {Q. \<exists>n::nat. Q = iter P n}" by (simp add: Collect_mono)
  have "P\<^sup>\<star>;P = \<Union>{R. \<exists>n::nat. R = iter P n};P" by (simp add: kleene_star_def)
  also have "\<dots> = \<Union>{Q. \<exists>R\<in>{R. \<exists>n::nat. R = iter P n}. Q = R;P}" by (simp add: lang_concat_distr_Union1)
  also have "\<dots> = \<Union>{Q. \<exists>R n::nat. R = iter P n \<and> Q = R;P}" by simp
  also have "\<dots> = \<Union>{Q. \<exists>n::nat. Q = (iter P n);P}" by auto
  also have "\<dots> = \<Union>{Q. \<exists>n::nat. Q = iter P (n+1)}" by (simp add: iteration_helper)
  also from a have "\<dots> \<subseteq> \<Union>{Q. \<exists>n::nat. Q = iter P n}" by (rule Union_mono)
  also have "\<dots> = P\<^sup>\<star>" by (simp add: kleene_star_def)
  finally show ?thesis .
qed

lemma kleene3: "skip \<union> P\<^sup>\<star>;P \<subseteq> P\<^sup>\<star>"
  by (metis skip_included_in_kleene_star kleene3part le_sup_iff)

lemma kleene4: "P \<union> (R;Q) \<subseteq> R \<Longrightarrow> P;Q\<^sup>\<star> \<subseteq> R"
proof -
  assume a: "P \<union> (R;Q) \<subseteq> R"
  have b: "\<forall>n::nat. P;(iter Q n) \<subseteq> R"
  proof (rule allI)
    fix n
    show "P;(iter Q n) \<subseteq> R"
    proof (induct n)
      have "P;(iter Q 0) = P" by (metis Languages.iter.simps(1) lang_concat_unit2)
      then show "P;(iter Q 0) \<subseteq> R" by (metis a le_supE)
    next
      fix n
      assume c: "P;(iter Q n) \<subseteq> R"
      have d: "P;(iter Q (Suc n)) = (P;(iter Q n));Q" by (metis Languages.iter.simps(2) lang_concat_associative iteration_helper)
      also have e: "\<dots> \<subseteq> R" by (metis (no_types) Un_upper2 a c lang_concat_monotone1 xt1(6))
      show "P;(iter Q (Suc n)) \<subseteq> R" by (metis calculation e)
    qed
  qed
  have "\<forall>T. (\<exists>n::nat. T = P;(iter Q n)) \<longrightarrow> T \<subseteq> R"
  proof (rule allI)
    fix T
    show "(\<exists>n. T = P;(iter Q n)) \<longrightarrow> T \<subseteq> R"
    proof (rule impI)
      assume "\<exists>n. T = P ; iter Q n"
      then obtain n where f: "T = P ; iter Q n" by blast
      from b f show "T \<subseteq> R" by metis
    qed
  qed
  then have c: "{T. \<exists>n::nat. T = P;(iter Q n)} \<subseteq> {T. T \<subseteq> R}" by (simp add: Collect_mono)
  have "P;Q\<^sup>\<star> = P;\<Union>{S. \<exists>n::nat. S = iter Q n}" by (simp add: kleene_star_def)
  also have "\<dots> = \<Union>{T. \<exists>S\<in>{S. \<exists>n::nat. S = iter Q n}. T = P;S}" by (simp add: lang_concat_distr_Union2)
  also have "\<dots> = \<Union>{T. \<exists>S n::nat. S = iter Q n \<and> T = P;S}" by simp
  also have "\<dots> = \<Union>{T. \<exists>n::nat. T = P;(iter Q n)}" by auto
  also from c have "\<dots> \<subseteq> \<Union>{T. T \<subseteq> R}" by (rule Union_mono)
  also have "\<dots> \<subseteq> R" by (metis Pow_def Union_Pow_eq subset_refl)
  finally show "P;Q\<^sup>\<star> \<subseteq> R" .
qed

lemma iter_monotone: "P \<subseteq> Q \<Longrightarrow> iter P n \<subseteq> iter Q n"
proof (induct n, simp, simp)
  fix n
  assume IH: "iter P n \<subseteq> iter Q n"
  assume "P \<subseteq> Q"
  hence "P ; iter P n \<subseteq> Q ; iter P n" by (metis lang_concat_monotone1)
  also from IH have "Q ; iter P n \<subseteq> Q ; iter Q n" by (metis lang_concat_monotone2)
  finally show "P ; iter P n \<subseteq> Q ; iter Q n" by metis
qed

lemma kleene_star_monotone: "P \<subseteq> Q \<Longrightarrow> P\<^sup>\<star> \<subseteq> Q\<^sup>\<star>"
  by (metis elem_of_kleene_star iter_monotone subsetD subsetI)

(* The exchange family. *)

lemma exchange: "(P\<parallel>Q);(R\<parallel>S) \<subseteq> (P;R)\<parallel>(Q;S)"
proof -
  have alternative: "\<forall>t P Q R S. t \<in> (P\<parallel>Q);(R\<parallel>S) \<longrightarrow>  t \<in> (P;R)\<parallel>(Q;S)"
  proof (rule allI)
    fix t
    show "\<forall>P Q R S. t \<in> (P\<parallel>Q);(R\<parallel>S) \<longrightarrow> t \<in> (P;R)\<parallel>(Q;S)"
    proof (induct t)
      show "\<forall>P Q R S. [] \<in> (P\<parallel>Q);(R\<parallel>S) \<longrightarrow> [] \<in> (P;R)\<parallel>(Q;S)"
      proof (rule allI, rule allI, rule allI, rule allI)
        fix P Q R S
        show "[] \<in> (P\<parallel>Q);(R\<parallel>S) \<longrightarrow> [] \<in> (P;R)\<parallel>(Q;S)" by (simp add: empty_elem_of_lang_concat empty_elem_of_lang_interleave)
      qed
    next
      fix e t
      assume IH: "\<forall>P Q R S. t \<in> (P\<parallel>Q);(R\<parallel>S) \<longrightarrow> t \<in> (P;R)\<parallel>(Q;S)"
      show "\<forall>P Q R S. e#t \<in> (P\<parallel>Q);(R\<parallel>S) \<longrightarrow> e#t \<in> (P;R)\<parallel>(Q;S)"
      proof (rule allI, rule allI, rule allI, rule allI)
        fix P Q R S
        show "e#t \<in> (P\<parallel>Q);(R\<parallel>S) \<longrightarrow> e#t \<in> (P;R)\<parallel>(Q;S)"
        proof (rule impI)
          assume "e#t \<in> (P\<parallel>Q);(R\<parallel>S)"
          then have "([] \<in> P\<parallel>Q \<and> e#t \<in> R\<parallel>S) \<or> (\<exists>t1 t2'. t = t1@t2' \<and> e#t1 \<in> P\<parallel>Q \<and> t2' \<in> R\<parallel>S)" by (simp add: nonempty_elem_of_lang_concat)
          moreover {
            assume "[] \<in> P\<parallel>Q \<and> e#t \<in> R\<parallel>S"
            then have "[] \<in> P \<and> [] \<in> Q \<and> e#t \<in> R\<parallel>S" by (simp add: empty_elem_of_lang_interleave)
            then have "R \<subseteq> P;R \<and> S \<subseteq> Q;S \<and> e#t \<in> R\<parallel>S" by (metis append_Nil elem_of_lang_concat subsetI)
            then have "e#t \<in> (P;R)\<parallel>(Q;S)" by (metis lang_interleave_monotone1 lang_interleave_monotone2 subsetD)
          } moreover {
            assume "\<exists>t1 t2'. t = t1@t2' \<and> e#t1 \<in> P\<parallel>Q \<and> t2' \<in> R\<parallel>S"
            then obtain t1 t2' where a: "t = t1@t2' \<and> t2' \<in> R\<parallel>S" and b: "e#t1 \<in> P\<parallel>Q" by blast
            from b have "(\<exists>p' q. e#p' \<in> P \<and> q \<in> Q \<and> t1 \<in> p'\<otimes>q) \<or> (\<exists>p q'. p \<in> P \<and> e#q' \<in> Q \<and> t1 \<in> p\<otimes>q')" by (simp add: nonempty_elem_of_lang_interleave)
            moreover {
              assume "\<exists>p' q. e#p' \<in> P \<and> q \<in> Q \<and> t1 \<in> p'\<otimes>q"
              then obtain p' q where c: "e#p' \<in> P \<and> q \<in> Q \<and> t1 \<in> p'\<otimes>q" by blast
              from c have d: "t1 \<in> {p'}\<parallel>{q}" by (metis elem_of_lang_interleave insertI1)
              from a d IH have e: "t \<in> ({p'};R)\<parallel>({q};S)" by (metis elem_of_lang_concat)
              then have f: "e#t \<in> ({e#p'};R)\<parallel>({q};S)"
              proof -
                from e have "\<exists>u1 u2. u1 \<in> {p'};R \<and> u2 \<in> {q};S \<and> t \<in> u1\<otimes>u2" by (simp add: elem_of_lang_interleave)
                then have "\<exists>u1 u2. u1 \<in> {p'};R \<and> u2 \<in> {q};S \<and> e#t \<in> (e#u1)\<otimes>u2" by (metis interleave_contains_nonempty)
                then have "\<exists>u1 u2. e#u1 \<in> {e#p'};R \<and> u2 \<in> {q};S \<and> e#t \<in> (e#u1)\<otimes>u2" by (metis append_Cons elem_of_lang_concat insertI1 singletonE)
                then show "e#t \<in> ({e#p'};R)\<parallel>({q};S)" by (metis elem_of_lang_interleave)
              qed
              have "e#t \<in> (P;R)\<parallel>(Q;S)"
              proof -
                from c have "{e#p'} \<subseteq> P \<and> {q} \<subseteq> Q" by (metis empty_subsetI insert_subset)
                then have "{e#p'};R \<subseteq> P;R \<and> {q};S \<subseteq> Q;S" by (metis lang_concat_monotone1)
                then have "({e#p'};R)\<parallel>({q};S) \<subseteq> (P;R)\<parallel>({q};S) \<and> (P;R)\<parallel>({q};S) \<subseteq> (P;R)\<parallel>(Q;S)" by (metis lang_interleave_monotone1 lang_interleave_monotone2)
                then have g: "({e#p'};R)\<parallel>({q};S) \<subseteq> (P;R)\<parallel>(Q;S)" by auto
                from f g show "e#t \<in> (P;R)\<parallel>(Q;S)" by auto
              qed
            } moreover {
              assume "\<exists>p q'. p \<in> P \<and> e#q' \<in> Q \<and> t1 \<in> p\<otimes>q'"
              then obtain p q' where c: "p \<in> P \<and> e#q' \<in> Q \<and> t1 \<in> p\<otimes>q'" by blast
              from c have d: "t1 \<in> {p}\<parallel>{q'}" by (metis elem_of_lang_interleave insertI1)
              from a d IH have e: "t \<in> ({p};R)\<parallel>({q'};S)" by (metis elem_of_lang_concat)
              then have f: "e#t \<in> ({p};R)\<parallel>({e#q'};S)"
              proof -
                from e have "\<exists>u1 u2. u1 \<in> {p};R \<and> u2 \<in> {q'};S \<and> t \<in> u1\<otimes>u2" by (simp add: elem_of_lang_interleave)
                then have "\<exists>u1 u2. u1 \<in> {p};R \<and> u2 \<in> {q'};S \<and> e#t \<in> u1\<otimes>(e#u2)" by (metis interleave_contains_nonempty)
                then have "\<exists>u1 u2. u1 \<in> {p};R \<and> e#u2 \<in> {e#q'};S \<and> e#t \<in> u1\<otimes>(e#u2)" by (metis append_Cons elem_of_lang_concat insertI1 singletonE)
                then show "e#t \<in> ({p};R)\<parallel>({e#q'};S)" by (metis elem_of_lang_interleave)
              qed
              have "e#t \<in> (P;R)\<parallel>(Q;S)"
              proof -
                from c have "{p} \<subseteq> P \<and> {e#q'} \<subseteq> Q" by (metis empty_subsetI insert_subset)
                then have "{p};R \<subseteq> P;R \<and> {e#q'};S \<subseteq> Q;S" by (metis lang_concat_monotone1)
                then have "({p};R)\<parallel>({e#q'};S) \<subseteq> (P;R)\<parallel>({e#q'};S) \<and> (P;R)\<parallel>({e#q'};S) \<subseteq> (P;R)\<parallel>(Q;S)" by (metis lang_interleave_monotone1 lang_interleave_monotone2)
                then have g: "({p};R)\<parallel>({e#q'};S) \<subseteq> (P;R)\<parallel>(Q;S)" by auto
                from f g show "e#t \<in> (P;R)\<parallel>(Q;S)" by auto
              qed
            }
            ultimately have "e#t \<in> (P;R)\<parallel>(Q;S)" by (rule disjE)
          }
          ultimately show "e#t \<in> (P;R)\<parallel>(Q;S)" by (rule disjE)
        qed
      qed
    qed
  qed
  then show ?thesis by (metis subsetI)
qed

lemma small_exchange1: "P;(Q\<parallel>R) \<subseteq> (P;Q)\<parallel>R"
  by (metis exchange lang_concat_unit1 lang_interleave_unit2)

lemma small_exchange2: "(P\<parallel>Q);R \<subseteq> P\<parallel>(Q;R)"
  by (metis exchange lang_concat_unit2 lang_interleave_unit1)

lemma interleave_contains_concat: "P;Q \<subseteq> P\<parallel>Q"
  by (metis lang_concat_unit2 lang_interleave_unit1 small_exchange1)

end
