(*  Title:       Recursion.thy
    Author:      Stephan van Staden, ETH Zurich, Switzerland
    Description: A treatment of recursion in terms of least fixpoints; deductive and operational rules for recursion.
    Comment:     See Inductive.thy for theorems about lfp.
    Copyright    2012 ETH Zurich
*)

theory Recursion
  imports DeductiveCalculi OperationalCalculi
begin

declare [[ smt_solver = remote_z3]]
declare [[ smt_timeout = 60 ]]
declare [[ z3_options = "-memory:500" ]]

(* Deductive rules. *)

locale deductive_recursion_rules = 
  fixes triple :: "'a \<Rightarrow> 'b::complete_lattice \<Rightarrow> 'a \<Rightarrow> bool"
  fixes Assertions :: "'a set"
  assumes choice: "a \<in> Assertions \<Longrightarrow> a' \<in> Assertions \<Longrightarrow> (\<forall>P \<in> X. triple a P a') \<Longrightarrow> triple a (Sup X) a'"
  assumes triple_downward_closed: "triple a P a' \<Longrightarrow> less_eq P' P \<Longrightarrow> triple a P' a'"
begin
definition specstmt :: "'a \<Rightarrow> 'a \<Rightarrow> 'b" where
  "specstmt a a' \<equiv> Sup {P. triple a P a'}"
lemma specstmt_triple_relationship: "a \<in> Assertions \<Longrightarrow> a' \<in> Assertions \<Longrightarrow> triple a P a' \<longleftrightarrow> less_eq P (specstmt a a')"
proof -
  assume a1: "a \<in> Assertions" and a2: "a' \<in> Assertions"
  show "triple a P a' \<longleftrightarrow> less_eq P (specstmt a a')"
  proof (rule iffI)
    assume "triple a P a'"
    thus "less_eq P (specstmt a a')" by (metis Collect_def Sup_le_iff mem_def order_refl specstmt_def)
  next
    have a3: "triple a (specstmt a a') a'"
    proof -
      have "\<forall>P' \<in> {P'. triple a P' a'}. triple a P' a'" by auto
      with a1 a2 show "triple a (specstmt a a') a'" by (metis choice specstmt_def)
    qed
    assume "less_eq P (specstmt a a')"
    with a3 triple_downward_closed show "triple a P a'" by metis
  qed
qed
lemma rec: "a \<in> Assertions \<Longrightarrow> a' \<in> Assertions \<Longrightarrow> (\<forall>P. triple a P a' \<longrightarrow> triple a (f(P)) a') \<Longrightarrow> triple a (lfp f) a'"
proof -
  assume a1: "a \<in> Assertions" and a2: "a' \<in> Assertions"
  assume "\<forall>P. triple a P a' \<longrightarrow> triple a (f(P)) a'"
  with a1 a2 specstmt_triple_relationship have "triple a (f(specstmt a a')) a'" by (metis order_refl)
  hence "less_eq (f(specstmt a a')) (specstmt a a')" by (metis a1 a2 specstmt_triple_relationship)
  hence "less_eq (lfp f) (specstmt a a')" by (metis lfp_lowerbound)
  thus "triple a (lfp f) a'" by (metis a1 a2 specstmt_triple_relationship)
qed
end

lemma Hrefinement: "P\<lbrace>Q\<rbrace>R \<Longrightarrow> Q' \<subseteq> Q \<Longrightarrow> P\<lbrace>Q'\<rbrace>R"
  by (unfold hoare_def, metis lang_concat_monotone2 subset_trans)

interpretation H: deductive_recursion_rules hoare "{P. True}"
apply (rule deductive_recursion_rules.intro)
apply (metis Hchoice)
by (metis Hrefinement)

lemma Hrec: "(\<forall>Q. P\<lbrace>Q\<rbrace>R \<longrightarrow> P\<lbrace>f(Q)\<rbrace>R) \<Longrightarrow> P\<lbrace>lfp f\<rbrace>R"
  by (auto simp add: H.rec)

lemma (in basic_views_calculus) Brefinement: "v \<lless>P\<ggreater> v' \<Longrightarrow> P' \<subseteq> P \<Longrightarrow> v \<lless>P'\<ggreater> v'"
  by (unfold b_triple_def, rule Hrefinement)

sublocale basic_views_calculus < B!:deductive_recursion_rules b_triple Views
apply (rule deductive_recursion_rules.intro)
apply (metis Bchoice)
by (metis Brefinement)

lemma (in basic_views_calculus) Brec: "v \<in> Views \<Longrightarrow> v' \<in> Views \<Longrightarrow> (\<forall>P. v \<lless>P\<ggreater> v' \<longrightarrow> v \<lless>f(P)\<ggreater> v') \<Longrightarrow> v \<lless>complete_lattice_class.lfp f\<ggreater> v'"
  by (metis B.rec)

lemma (in framing_calculus) Frefinement: "v \<langle>P\<rangle> v' \<Longrightarrow> P' \<subseteq> P \<Longrightarrow> v \<langle>P'\<rangle> v'"
  by (unfold f_triple_def, metis Brefinement)

sublocale framing_calculus < F!:deductive_recursion_rules f_triple Views
apply (rule deductive_recursion_rules.intro)
apply (metis Fchoice)
by (metis Frefinement)

lemma (in framing_calculus) Frec: "v \<in> Views \<Longrightarrow> v' \<in> Views \<Longrightarrow> (\<forall>P. v \<langle>P\<rangle> v' \<longrightarrow> v \<langle>f(P)\<rangle> v') \<Longrightarrow> v \<langle>complete_lattice_class.lfp f\<rangle> v'"
  by (metis F.rec)

lemma (in full_views_calculus) Vrefinement: "\<lbrace>v\<rbrace> C \<lbrace>v'\<rbrace> \<Longrightarrow> C' \<subseteq> C \<Longrightarrow> \<lbrace>v\<rbrace> C' \<lbrace>v'\<rbrace>"
  by (unfold views_triple_def, metis subsetD)

sublocale full_views_calculus < V!:deductive_recursion_rules views_triple Views
apply (rule deductive_recursion_rules.intro)
apply (metis Vchoice)
by (metis Vrefinement)

lemma (in full_views_calculus) Vrec: "v \<in> Views \<Longrightarrow> v' \<in> Views \<Longrightarrow> (\<forall>C. \<lbrace>v\<rbrace> C \<lbrace>v'\<rbrace> \<longrightarrow> \<lbrace>v\<rbrace> f(C) \<lbrace>v'\<rbrace>) \<Longrightarrow> \<lbrace>v\<rbrace> complete_lattice_class.lfp f \<lbrace>v'\<rbrace>"
  by (metis V.rec)


(* Operational rules. *)

context abstract_small_step_calculi
begin
lemma Prec: "mono f \<Longrightarrow> <f(lfp f),s>\<rightarrow>\<^isub>a<P,s'> \<Longrightarrow> <lfp f,s>\<rightarrow>\<^isub>a<P,s'>"
  by (metis lfp_unfold)
lemma Prec': "mono f \<Longrightarrow> <lfp f,s>\<rightarrow>\<^isub>a<f(lfp f),s>"
  by (metis a_plotkin_def lang_concat_unit1 lang_concat_unit2 lfp_unfold set_eq_subset skip_in_Actions)
lemma Mrec: "mono f \<Longrightarrow> f(lfp f) -P\<rightarrow>\<^isub>a P' \<Longrightarrow> (lfp f) -P\<rightarrow>\<^isub>a P'"
  by (metis lfp_unfold)
lemma Mrec': "mono f \<Longrightarrow> (lfp f) -skip\<rightarrow>\<^isub>a f(lfp f)"
  by (metis a_milner_def lang_concat_unit1 lfp_lemma2 skip_in_Actions)
end
lemma Krec: "mono f \<Longrightarrow> <f(lfp f),s>\<rightarrow>\<^isub>as' \<Longrightarrow> <lfp f,s>\<rightarrow>\<^isub>as'"
  by (metis lfp_unfold)

context small_step_calculi_for_descriptions
begin
lemma PDrec: "mono f \<Longrightarrow> <f(lfp f),\<sigma>>\<rightarrow>\<^isub>d<P,\<sigma>'> \<Longrightarrow> <lfp f,\<sigma>>\<rightarrow>\<^isub>d<P,\<sigma>'>"
  by (metis lfp_unfold)
lemma PDrec': "mono f \<Longrightarrow> <lfp f,\<sigma>>\<rightarrow>\<^isub>d<f(lfp f),\<sigma>>"
  by (unfold d_plotkin_def, metis a.Prec' ic_traces_ending_in_state_not_empty all_not_in_conv)
lemma MDrec: "mono f \<Longrightarrow> f(lfp f) -P\<rightarrow>\<^isub>a P' \<Longrightarrow> (lfp f) -P\<rightarrow>\<^isub>a P'"
  by (metis lfp_unfold)
lemma MDrec': "mono f \<Longrightarrow> (lfp f) -skip\<rightarrow>\<^isub>a f(lfp f)"
  by (metis a.Mrec')
end
lemma KDrec: "mono f \<Longrightarrow> <f(lfp f),\<sigma>>\<rightarrow>\<^isub>d\<sigma>' \<Longrightarrow> <lfp f,\<sigma>>\<rightarrow>\<^isub>d\<sigma>'"
  by (metis lfp_unfold)

context small_step_calculi_for_commands
begin
lemma PCrec: "mono f \<Longrightarrow> <f(lfp f),\<sigma>>\<rightarrow>\<^isub>c<C,\<sigma>'> \<Longrightarrow> <lfp f,\<sigma>>\<rightarrow>\<^isub>c<C,\<sigma>'>"
  by (metis lfp_unfold)
lemma PCrec': "mono f \<Longrightarrow> <lfp f,\<sigma>>\<rightarrow>\<^isub>c<f(lfp f),\<sigma>>"
  by (metis PCseq1 lang_concat_unit1 lfp_unfold)
lemma MCrec: "mono f \<Longrightarrow> f(lfp f) -C\<rightarrow>\<^isub>c C' \<Longrightarrow> (lfp f) -C\<rightarrow>\<^isub>c C'"
  by (metis lfp_unfold)
lemma MCrec': "mono f \<Longrightarrow> (lfp f) -skip\<rightarrow>\<^isub>c f(lfp f)"
  by (metis MCseq1 lang_concat_unit1 lfp_unfold)
end
lemma KCrec: "mono f \<Longrightarrow> <f(lfp f),\<sigma>>\<rightarrow>\<^isub>c\<sigma>' \<Longrightarrow> <lfp f,\<sigma>>\<rightarrow>\<^isub>c\<sigma>'"
  by (metis lfp_unfold)

(* Relating fixpoints *)

definition galois_connection :: "('a::order \<Rightarrow> 'b::order) \<Rightarrow> ('b::order \<Rightarrow> 'a::order) \<Rightarrow> bool" where
  "galois_connection f g \<equiv> (\<forall>x y. f x \<le> y \<longleftrightarrow> x \<le> g y)"

lemma galois_deflation: "galois_connection f g \<Longrightarrow> f (g x) \<le> x"
  by (metis galois_connection_def order_refl)

definition Sup_distr :: "('a::complete_lattice \<Rightarrow> 'b::complete_lattice) \<Rightarrow> bool" where
  "Sup_distr f \<equiv> \<forall>X. f(Sup X) = Sup (f ` X)"

lemma Sup_distr_is_mono:
  assumes Sup_distr_f: "Sup_distr f"
  shows "mono f"
proof (unfold mono_def, rule allI, rule allI, rule impI)
  fix x::'a fix y
  assume "x \<le> y"
  hence "Sup {x,y} = y" by (metis (no_types) Sup_empty Sup_insert sup_absorb1 sup_bot_left sup_commute)
  hence "f y = f (Sup {x,y})" by auto
  also have "\<dots> = Sup (f ` {x,y})" by (metis Sup_distr_f Sup_distr_def)
  also have "\<dots> = Sup {f x, f y}" by auto
  ultimately show "f x \<le> f y" by (metis Sup_le_iff insertI1 order_refl)
qed

lemma Sup_distr_implies_lower_adjoint:
  assumes Sup_distr_f: "Sup_distr f"
  shows "galois_connection f (\<lambda>y. Sup {x. f x \<le> y})"
proof (unfold galois_connection_def, rule allI, rule allI, rule iffI)
  fix x y
  assume "f x \<le> y"
  thus "x \<le> Sup {x. f x \<le> y}" by (metis Collect_def Sup_upper mem_def)
next
  fix x y
  from Sup_distr_f have mono_f: "mono f" by (rule Sup_distr_is_mono)
  assume "x \<le> Sup {x. f x \<le> y}"
  hence "f x \<le> f (Sup {x. f x \<le> y})" by (metis mono_f mono_def)
  also have "\<dots> = Sup (f ` {x. f x \<le> y})" by (metis Sup_distr_f Sup_distr_def)
  also have "\<dots> \<le> y" by (auto simp add: Collect_def Sup_le_iff imageE mem_def)
  ultimately show "f x \<le> y" by simp
qed

lemma lfp_fusion:
  assumes Sup_distr_f: "Sup_distr f"
  and mono_h: "mono h" and mono_k: "mono k"
  and distr: "f \<circ> h = k \<circ> f"
  shows "f (lfp h) = lfp k"
proof -
  have a: "k (f (lfp h)) = f (lfp h)" by (metis distr lfp_unfold mono_h o_def)
  have b: "\<And>x. k x = x \<Longrightarrow> f (lfp h) \<le> x"
  proof -
    let ?g = "\<lambda>y. Sup {x. f x \<le> y}"
    from Sup_distr_f have galois: "galois_connection f ?g" by (rule Sup_distr_implies_lower_adjoint)
    fix x
    assume x_fixpoint_k: "k x = x"
    have "h (?g x) \<le> ?g x"
    proof -
      have "f (?g x) \<le> x" by (metis galois galois_deflation)
      hence "k (f (?g x)) \<le> x" by (metis Collect_def monoD mono_k x_fixpoint_k)
      hence "f (h (?g x)) \<le> x" by (metis distr o_def)
      thus "h (?g x) \<le> ?g x" by (metis Collect_def Sup_upper mem_def)
    qed
    thus "f (lfp h) \<le> x" by (metis galois galois_connection_def lfp_lowerbound)
  qed
  from a b show "f (lfp h) = lfp k" by (metis lfp_lowerbound lfp_unfold mono_k order_antisym)
qed

lemma Sup_distr_trace_set_of_command: "Sup_distr trace_set_of_command"
  by (unfold Sup_distr_def, rule allI, auto simp add: choice_homomorphism)
lemma trace_set_of_command_fusion:
  assumes mono_f: "mono f" and mono_g: "mono g"
  and distr: "trace_set_of_command \<circ> f = g \<circ> trace_set_of_command"
  shows "trace_set_of_command (lfp f) = lfp g"
by (metis Sup_distr_trace_set_of_command distr lfp_fusion mono_f mono_g)

lemma (in order) mono_closed_o: "mono f \<Longrightarrow> mono f' \<Longrightarrow> mono (f \<circ> f')"
  by (unfold mono_def, metis o_eq_dest_lhs order_class.monoD)

locale trace_set_of_command_fusion_example
begin
lemma example_function_mono: "mono (\<lambda>x. skip \<union> P\<parallel>x)"
proof -
  have a1: "mono (\<lambda>y. skip \<union> y)" by (metis lang_union_monotone2 monoI)
  have a2: "mono (\<lambda>x. P\<parallel>x)" by (metis lang_interleave_monotone2 monoI)
  from a1 a2 have "mono ((\<lambda>y. skip \<union> y) \<circ> (\<lambda>x. P\<parallel>x))" by (rule mono_closed_o)
  thus "mono (\<lambda>x. skip \<union> P\<parallel>x)" by (simp add: o_def)
qed
lemma example_distr: "trace_set_of_command \<circ> (\<lambda>x. skip \<union> C\<parallel>x) = (\<lambda>x. skip \<union> trace_set_of_command(C)\<parallel>x) \<circ> trace_set_of_command" 
proof -
  have "trace_set_of_command \<circ> (\<lambda>x. skip \<union> C\<parallel>x) = (\<lambda>x. trace_set_of_command(skip) \<union> trace_set_of_command(C\<parallel>x))" by (metis binary_choice_homomorphism o_apply)
  also have "\<dots> = (\<lambda>x. skip \<union> trace_set_of_command(C)\<parallel>trace_set_of_command(x))" by (metis skip_homomorphism conc_homomorphism)
  also have "\<dots> = (\<lambda>x. skip \<union> trace_set_of_command(C)\<parallel>x) \<circ> trace_set_of_command" by auto
  finally show ?thesis by simp
qed
lemma trace_set_of_command_fusion_example: "trace_set_of_command (lfp (\<lambda>x. skip \<union> C\<parallel>x)) = lfp (\<lambda>x. skip \<union> trace_set_of_command(C)\<parallel>x)"
  by (metis example_distr example_function_mono trace_set_of_command_fusion)
end

(* Further remarks. *)

lemma iteration_special_case_of_recursion: "P\<^sup>\<star> = lfp (\<lambda>x. skip \<union> P;x)"
proof -
  have part1: "P\<^sup>\<star> \<supseteq> lfp (\<lambda>x. skip \<union> P;x)" by (metis kleene1 lfp_lowerbound)
  have a: "mono (\<lambda>x. skip \<union> P;x)"
  proof -
    have a1: "mono (\<lambda>y. skip \<union> y)" by (metis lang_union_monotone2 monoI)
    have a2: "mono (\<lambda>y. P;y)" by (metis lang_concat_monotone2 monoI)
    from a1 a2 have "mono ((\<lambda>y. skip \<union> y) \<circ> (\<lambda>y. P;y))" by (rule mono_closed_o)
    thus "mono (\<lambda>x. skip \<union> P;x)" by (simp add: o_def)
  qed
  have b: "\<And>P'. skip \<union> P;P' \<subseteq> P' \<Longrightarrow> P\<^sup>\<star> \<subseteq> P'" by (metis kleene2 lang_concat_unit2)
  have part2: "P\<^sup>\<star> \<subseteq> lfp (\<lambda>x. skip \<union> P;x)" by (metis (full_types) a b lfp_unfold subset_refl)
  from part1 part2 show ?thesis by simp
qed


(* The Kleene fixed point theorem can be generalized to complete partial orders, but complete lattices offer sufficient generality for formal languages (including descriptions and commands, of course). *)

definition (in order) directed where
  "directed X \<equiv> X \<noteq> {} \<and> (\<forall>P \<in> X. \<forall>Q \<in> X. \<exists>R \<in> X. less_eq P R \<and> less_eq Q R)"
definition Scott_continuous :: "('a::complete_lattice \<Rightarrow> 'b::complete_lattice) \<Rightarrow> bool" where
  "Scott_continuous f \<equiv> \<forall>X. directed(X) \<longrightarrow> f(Sup X) = Sup {Q. \<exists>P \<in> X. Q = f(P)}"
lemma Scott_continuous_implies_mono: "Scott_continuous f \<Longrightarrow> mono f"
proof -
  assume "Scott_continuous f"
  hence a1: "\<forall>X. directed(X) \<longrightarrow> f(Sup X) = Sup {Q. \<exists>P \<in> X. Q = f(P)}" by (unfold Scott_continuous_def)
  have "\<And>P Q. less_eq (P::'a) Q \<Longrightarrow> less_eq (f P) (f Q)"
  proof - fix P Q
    assume "less_eq (P::'a) Q"
    hence a2: "directed {P, Q}" and a3: "Sup {P, Q} = Q" by (unfold directed_def, auto simp add: le_iff_sup)
    from a1 a2 have "f(Sup {P, Q}) = Sup {Q'. \<exists>P' \<in> {P, Q}. Q' = f(P')}" by auto
    with a3 have "f Q = Sup {Q'. \<exists>P' \<in> {P, Q}. Q' = f(P')}" by simp
    also have "\<dots> = Sup {Q'. Q' = f(P) \<or> Q' = f(Q)}" by auto
    also have "\<dots> = Sup {f P, f Q}" by (metis Collect_disj_eq insert_is_Un singleton_conv)
    also have "\<dots> = sup (f P) (f Q)" by (metis (mono_tags) Sup_insert Un_left_absorb calculation insert_is_Un)
    finally have "f Q = sup (f P) (f Q)" by simp
    thus "less_eq (f P) (f Q)" by (metis le_iff_sup)
  qed
  thus "mono f" by (metis monoI)
qed
fun funiter :: "('a \<Rightarrow> 'a) \<Rightarrow> nat \<Rightarrow> ('a \<Rightarrow> 'a)" where
   "funiter _ 0 = id"
 | "funiter f (Suc n) = f \<circ> (funiter f n)"
lemma Kleene_fixpoint_helper1: "Scott_continuous f \<Longrightarrow> funiter f n bot \<le> funiter f (n+m) bot"
proof -
  assume a: "Scott_continuous f"
  show "funiter f n bot \<le> funiter f (n+m) bot"
    by (induct n, simp, simp, metis a Scott_continuous_implies_mono mono_def)
qed
lemma Kleene_fixpoint_helper2: "Scott_continuous f \<Longrightarrow> funiter f n bot \<le> lfp f"
proof -
  assume a: "Scott_continuous f"
  show "funiter f n bot \<le> lfp f"
    by (induct n, simp, simp, metis a Scott_continuous_implies_mono mono_def lfp_unfold)
qed
theorem Kleene_fixpoint_theorem: "Scott_continuous f \<Longrightarrow> lfp f = Sup {P. \<exists>n. P = funiter f n bot}"
proof -
  assume a: "Scott_continuous f"
  have b: "directed {P. \<exists>n. P = funiter f n bot}"
  proof (unfold directed_def, rule conjI)
    show "{P. \<exists>n. P = funiter f n bot} \<noteq> {}" by auto
  next
    show "\<forall>P\<in>{P. \<exists>n. P = funiter f n bot}. \<forall>Q\<in>{P. \<exists>n. P = funiter f n bot}. \<exists>R\<in>{P. \<exists>n. P = funiter f n bot}. P \<le> R \<and> Q \<le> R"
    proof (rule ballI, rule ballI)
      fix P Q
      assume "P \<in> {P. \<exists>n. P = funiter f n bot}"
      hence "\<exists>n. P = funiter f n bot" by auto
      then obtain n where b1: "P = funiter f n bot" by blast
      assume "Q \<in> {P. \<exists>n. P = funiter f n bot}"
      hence "\<exists>m. Q = funiter f m bot" by auto
      then obtain m where b2: "Q = funiter f m bot" by blast
      let ?R = "funiter f (max n m) bot"
      have c1: "?R \<in> {P. \<exists>n. P = funiter f n bot}" by auto
      have c2: "P \<le> ?R \<and> Q \<le> ?R" by (metis a b1 b2 Kleene_fixpoint_helper1 comm_semiring_1_class.normalizing_semiring_rules(24) min_max.sup_commute nat_minus_add_max)
      from c1 c2 show "\<exists>R\<in>{P. \<exists>n. P = funiter f n bot}. P \<le> R \<and> Q \<le> R" by metis
    qed
  qed
  have c: "\<And>Q. (\<exists>P. \<exists>n. P = funiter f n bot \<and> Q = f P) \<longleftrightarrow> (\<exists>n. Q = f (funiter f n bot))" by auto
  from a have "f (Sup {P. \<exists>n. P = funiter f n bot}) = Sup {Q. \<exists>P \<in> {P. \<exists>n. P = funiter f n bot}. Q = f P}" by (unfold Scott_continuous_def, auto simp add: b)
  also have "\<dots> = Sup {Q. \<exists>P. \<exists>n. P = funiter f n bot \<and> Q = f P}" by auto
  also have "\<dots> = Sup {Q. \<exists>n. Q = f (funiter f n bot)}" by (simp only: c)
  also have "\<dots> = Sup {Q. \<exists>n. Q = funiter f (Suc n) bot}" by simp
  also have "\<dots> = Sup {Q. \<exists>n. Q = funiter f n bot}"
  proof -
    have d1: "\<And>Q. (\<exists>n. Q = funiter f n bot) \<longleftrightarrow> (\<exists>n. Q = funiter f (Suc n) bot) \<or> Q = funiter f 0 bot" by (metis nat.exhaust)
    have d2: "{Q. (\<exists>n. Q = funiter f (Suc n) bot) \<or> Q = bot} = {Q. \<exists>n. Q = funiter f (Suc n) bot} \<union> {bot}" by auto
    from d1 have "Sup {Q. \<exists>n. Q = funiter f n bot} = Sup {Q. (\<exists>n. Q = funiter f (Suc n) bot) \<or> Q = bot}" by auto
    also have "\<dots> = Sup ({Q. \<exists>n. Q = funiter f (Suc n) bot} \<union> {bot})" by (simp only: d2)
    also have "\<dots> = Sup {Q. \<exists>n. Q = funiter f (Suc n) bot}" by (simp add: Sup_union_distrib)
    finally show ?thesis by simp
  qed
  finally have fixpoint: "f (Sup {P. \<exists>n. P = funiter f n bot}) = Sup {P. \<exists>n. P = funiter f n bot}" by simp
  from a have "Sup {P. \<exists>n. P = funiter f n bot} \<le> lfp f" by (auto simp add: Collect_def Kleene_fixpoint_helper2 Sup_le_iff mem_def)
  with fixpoint show "lfp f = Sup {P. \<exists>n. P = funiter f n bot}" by (metis (no_types) lfp_lowerbound order_eq_iff)
qed
lemma Scott_continuous_closed_o: "Scott_continuous f \<Longrightarrow> Scott_continuous g \<Longrightarrow> Scott_continuous (g \<circ> f)"
proof -
  assume a1: "Scott_continuous f" and a2: "Scott_continuous g"
  show "Scott_continuous (g \<circ> f)"
  proof (unfold Scott_continuous_def, rule allI)
    fix X
    show "directed(X) \<longrightarrow> (g \<circ> f) (Sup X) = Sup {Q. \<exists>P\<in>X. Q = (g \<circ> f) P}"
    proof (rule impI)
      assume a3: "directed(X)"
      from a1 have b1: "f (Sup X) = Sup {Q. \<exists>P \<in> X. Q = f(P)}" by (unfold Scott_continuous_def, auto simp add: a3)
      have b2: "directed {Q. \<exists>P \<in> X. Q = f(P)}"
      proof (unfold directed_def, rule conjI)
        from a3 show "{Q. \<exists>P\<in>X. Q = f P} \<noteq> {}" by (unfold directed_def, auto)
      next
        show "\<forall>P\<in>{Q. \<exists>P\<in>X. Q = f P}. \<forall>Q\<in>{Q. \<exists>P\<in>X. Q = f P}. \<exists>R\<in>{Q. \<exists>P\<in>X. Q = f P}. P \<le> R \<and> Q \<le> R"
        proof (rule ballI, rule ballI) fix P Q
          assume "P \<in> {Q. \<exists>P\<in>X. Q = f P}"
          hence "\<exists>P'\<in>X. P = f P'" by auto
          then obtain P' where c1: "P' \<in> X" and c2: "P = f P'" by blast
          assume "Q \<in> {Q. \<exists>P\<in>X. Q = f P}"
          hence "\<exists>Q'\<in>X. Q = f Q'" by auto
          then obtain Q' where c3: "Q' \<in> X" and c4: "Q = f Q'" by blast
          from a3 c1 c3 have "\<exists>R \<in> X. P' \<le> R \<and> Q' \<le> R" by (metis directed_def)
          then obtain R where c5: "R \<in> X" and "P' \<le> R \<and> Q' \<le> R" by blast
          hence "f P' \<le> f R \<and> f Q' \<le> f R" by (metis a1 Scott_continuous_implies_mono mono_def)
          with c2 c4 c5 show "\<exists>R\<in>{Q. \<exists>P\<in>X. Q = f P}. P \<le> R \<and> Q \<le> R" by auto
        qed
      qed
      have "(g \<circ> f) (Sup X) = g (f (Sup X))" by auto
      also have "\<dots> = g (Sup {Q. \<exists>P \<in> X. Q = f(P)})" by (simp add: b1)
      also from a2 have "\<dots> = Sup {Q. \<exists>P' \<in> {P'. \<exists>P \<in> X. P' = f(P)}. Q = g P'}" by (unfold Scott_continuous_def, auto simp add: b2)
      also have "\<dots> = Sup {Q. \<exists>P'. \<exists>P \<in> X. P' = f(P) \<and> Q = g P'}" by auto
      also have "\<dots> = Sup {Q. \<exists>P \<in> X. Q = g (f P)}" by metis
      also have "\<dots> = Sup {Q. \<exists>P \<in> X. Q = (g \<circ> f) P}" by (simp add: o_def)
      finally show "(g \<circ> f) (Sup X) = Sup {Q. \<exists>P\<in>X. Q = (g \<circ> f) P}" by simp
    qed
  qed
qed
lemma distr_through_Sup_implies_Scott_continuous: "(\<forall>X. X \<noteq> {} \<longrightarrow> f(Sup X) = Sup {P. \<exists>Q \<in> X. P = f Q}) \<Longrightarrow> Scott_continuous f"
  by (auto simp add: directed_def Scott_continuous_def)

lemma lang_concat_Scott_cont_1: "Scott_continuous (\<lambda>x. x;P)"
  by (auto simp add: distr_through_Sup_implies_Scott_continuous lang_concat_distr_Union1)
lemma lang_concat_Scott_cont_2: "Scott_continuous (\<lambda>x. P;x)"
  by (auto simp add: distr_through_Sup_implies_Scott_continuous lang_concat_distr_Union2)
lemma lang_union_Scott_cont_1: "Scott_continuous (\<lambda>x. x\<union>P)"
  by (auto simp add: distr_through_Sup_implies_Scott_continuous lang_union_distr_Union1)
lemma lang_union_Scott_cont_2: "Scott_continuous (\<lambda>x. P\<union>x)"
  by (auto simp add: distr_through_Sup_implies_Scott_continuous lang_union_distr_Union2)
lemma lang_interleave_Scott_cont_1: "Scott_continuous (\<lambda>x. x\<parallel>P)"
  by (auto simp add: distr_through_Sup_implies_Scott_continuous lang_interleave_distr_Union1)
lemma lang_interleave_Scott_cont_2: "Scott_continuous (\<lambda>x. P\<parallel>x)"
  by (auto simp add: distr_through_Sup_implies_Scott_continuous lang_interleave_distr_Union2)

lemma iter_Scott_cont: "Scott_continuous (\<lambda>x. iter x n)"
proof (unfold Scott_continuous_def, rule allI, rule impI)
  fix X :: "'a language set"
  assume a: "directed X"
  hence b: "\<exists>P. P \<in> X" by (auto simp add: directed_def)
  show "iter (\<Union>X) n = \<Union>{Q. \<exists>P\<in>X. Q = iter P n}"
  proof (induct n, simp add: b, simp)
    fix n
    assume IH: "iter (\<Union>X) n = \<Union>{Q. \<exists>P\<in>X. Q = iter P n}"
    hence "\<Union>X ; \<Union>{Q. \<exists>P\<in>X. Q = iter P n} = \<Union>X ; iter (\<Union>X) n" by simp
    also have "\<dots> = \<Union>{Q. \<exists>P\<in>X. Q = P ; iter (\<Union>X) n}" by (metis lang_concat_distr_Union1)
    also have "\<dots> = \<Union>{Q. \<exists>P\<in>X. Q = P ; \<Union>{Q'. \<exists>P'\<in>X. Q' = iter P' n}}" by (simp add: IH)
    also have "\<dots> = \<Union>{Q. \<exists>P\<in>X. \<exists>P'\<in>X. Q = P ; iter P' n}" by (auto simp add: lang_concat_distr_Union2)
    also have "\<dots> = \<Union>{Q. \<exists>P\<in>X. Q = P ; iter P n}"
    proof (rule antisym)
      have "\<forall>q. q \<in> \<Union>{Q. \<exists>P\<in>X. \<exists>P'\<in>X. Q = P ; iter P' n} \<longrightarrow> q \<in> \<Union>{Q. \<exists>P\<in>X. Q = P ; iter P n}"
      proof (rule allI, rule impI)
        fix q
        assume "q \<in> \<Union>{Q. \<exists>P\<in>X. \<exists>P'\<in>X. Q = P ; iter P' n}"
        hence "\<exists>P\<in>X. \<exists>P'\<in>X. q \<in> P ; iter P' n" by auto
        then obtain P P' where c1: "P \<in> X" and c2: "P' \<in> X" and c3: "q \<in> P ; iter P' n" by blast
        from a directed_def c1 c2 have "\<exists>R \<in> X. P \<subseteq> R \<and> P' \<subseteq> R" by auto
        then obtain R where c4: "R \<in> X" and c5: "P \<subseteq> R \<and> P' \<subseteq> R" by blast
        from c3 c5 have "q \<in> R ; iter R n" by (metis lang_concat_monotone1 lang_concat_monotone2 iter_monotone set_rev_mp)
        with c4 show "q \<in> \<Union>{Q. \<exists>P\<in>X. Q = P ; iter P n}" by auto
      qed
      then show "\<Union>{Q. \<exists>P\<in>X. \<exists>P'\<in>X. Q = P ; iter P' n} \<subseteq> \<Union>{Q. \<exists>P\<in>X. Q = P ; iter P n}" by (simp only: subsetI)
    next
      show "\<Union>{Q. \<exists>P\<in>X. Q = P ; iter P n} \<subseteq> \<Union>{Q. \<exists>P\<in>X. \<exists>P'\<in>X. Q = P ; iter P' n}" by auto
    qed
    finally show "\<Union>X ; \<Union>{Q. \<exists>P\<in>X. Q = iter P n} = \<Union>{Q. \<exists>P\<in>X. Q = P ; iter P n}" by simp
  qed
qed

lemma kleene_star_Scott_cont: "Scott_continuous (\<lambda>x. x\<^sup>\<star>)"
proof (unfold kleene_star_def, unfold Scott_continuous_def, rule allI, rule impI)
  fix X :: "'a language set"
  assume a: "directed X"
  from iter_Scott_cont have "\<And>n. iter (\<Union>X) n = \<Union>{Q. \<exists>P\<in>X. Q = iter P n}"
  proof (unfold Scott_continuous_def)
    fix n
    assume "\<And>n. \<forall>X::'a language set. directed X \<longrightarrow> iter (\<Union>X) n = \<Union>{Q. \<exists>P\<in>X. Q = iter P n}"
    thus "iter (\<Union>X) n = \<Union>{Q. \<exists>P\<in>X. Q = iter P n}" by (auto simp add: a)
  qed
  then show "\<Union>{Q. \<exists>n. Q = iter (\<Union>X) n} = \<Union>{Q. \<exists>P\<in>X. Q = \<Union>{Q. \<exists>n. Q = iter P n}}" by auto
qed

end
